from decimal import Decimal

from babel.core import Locale
from django.conf import settings
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy
import babel
import babel.numbers
import easymoney
import floppyforms.__future__ as forms
import floppyforms.widgets


__all__ = (
    '_BaseMoneyInput',
    'CheckboxInput',
    'ClearableFileInput', 'ColorInput',
    '_CurrencyInput', 'DateInput', 'DateTimeInput', 'EmailInput', 'FileInput',
    'HiddenInput', 'IPAddressInput', 'Input', '_RealWorldCurrencyInput',
    'NullBooleanSelect', 'NumberInput', 'PasswordInput',
    'PhoneNumberInput', 'RadioSelect', 'RadioSelectHorizontal', 'RangeInput',
    'SearchInput', 'Select', 'SelectDateWidget',
    'SliderInput', 'SlugInput', 'SplitDateTimeWidget',
    'SplitHiddenDateTimeWidget', 'TextInput', 'Textarea', 'TimeInput',
    'URLInput', 'Widget',
)


Widget = floppyforms.widgets.Widget
Input = floppyforms.widgets.Input
TextInput = floppyforms.widgets.TextInput
PasswordInput = floppyforms.widgets.PasswordInput
HiddenInput = floppyforms.widgets.HiddenInput
SlugInput = floppyforms.widgets.SlugInput
IPAddressInput = floppyforms.widgets.IPAddressInput
FileInput = floppyforms.widgets.FileInput
ClearableFileInput = floppyforms.widgets.ClearableFileInput
Textarea = floppyforms.widgets.Textarea
DateInput = floppyforms.widgets.DateInput
DateTimeInput = floppyforms.widgets.DateTimeInput
TimeInput = floppyforms.widgets.TimeInput
SearchInput = floppyforms.widgets.SearchInput
EmailInput = floppyforms.widgets.EmailInput
URLInput = floppyforms.widgets.URLInput
ColorInput = floppyforms.widgets.ColorInput
NumberInput = floppyforms.widgets.NumberInput
RangeInput = floppyforms.widgets.RangeInput
PhoneNumberInput = floppyforms.widgets.PhoneNumberInput
CheckboxInput = floppyforms.widgets.CheckboxInput
Select = floppyforms.widgets.Select
NullBooleanSelect = floppyforms.widgets.NullBooleanSelect
RadioSelect = floppyforms.widgets.RadioSelect
SplitDateTimeWidget = floppyforms.widgets.SplitDateTimeWidget
SplitHiddenDateTimeWidget = floppyforms.widgets.SplitHiddenDateTimeWidget
SelectDateWidget = floppyforms.widgets.SelectDateWidget

# don't use Multiple widgets because they don't correspond to any model field
# CheckboxSelectMultiple = floppyforms.widgets.CheckboxSelectMultiple
# MultiWidget = floppyforms.widgets.MultiWidget
# SelectMultiple = floppyforms.widgets.SelectMultiple
# MultipleHiddenInput = floppyforms.widgets.MultipleHiddenInput
# class CheckboxSelectMultipleHorizontal(forms.CheckboxSelectMultiple):
#    template_name = 'floppyforms/checkbox_select_horizontal.html'


class _BaseMoneyInput(forms.NumberInput):
    # step = 0.01
    template_name = 'floppyforms/moneyinput.html'

    def get_context(self, *args, **kwargs):
        context = super(_BaseMoneyInput, self).get_context(*args, **kwargs)
        context['currency_symbol'] = self.CURRENCY_SYMBOL
        context['currency_symbol_is_prefix'] = self.currency_symbol_is_prefix
        return context

    if settings.USE_POINTS:
        currency_symbol_is_prefix = False
    else:
        locale = Locale.parse(settings.REAL_WORLD_CURRENCY_LOCALE)
        pattern = locale.currency_formats['standard']
        currency_symbol_is_prefix = u'\xa4' in pattern.prefix[0]

    def _format_value(self, value):
        if isinstance(value, easymoney.Money):
            value = Decimal(value)
        return force_text(value)


class _RealWorldCurrencyInput(_BaseMoneyInput):
    CURRENCY_SYMBOL = babel.numbers.get_currency_symbol(
        settings.REAL_WORLD_CURRENCY_CODE,
        settings.REAL_WORLD_CURRENCY_LOCALE
    )


class _CurrencyInput(_RealWorldCurrencyInput):
    if settings.USE_POINTS:
        if hasattr(settings, 'POINTS_CUSTOM_NAME'):
            CURRENCY_SYMBOL = settings.POINTS_CUSTOM_NAME
        else:
            # Translators: the label next to a "points" input field
            CURRENCY_SYMBOL = ugettext_lazy('points')


class RadioSelectHorizontal(forms.RadioSelect):
    template_name = 'floppyforms/radio_select_horizontal.html'


class SliderInput(forms.RangeInput):
    template_name = 'floppyforms/slider.html'
    show_value = True

    def __init__(self, *args, **kwargs):
        show_value = kwargs.pop('show_value', None)
        if show_value is not None:
            self.show_value = show_value
        super(SliderInput, self).__init__(*args, **kwargs)

    def _format_value(self, value):
        if isinstance(value, easymoney.Money):
            value = Decimal(value)
        return force_text(value)

    def get_context(self, *args, **kwargs):
        context = super(SliderInput, self).get_context(*args, **kwargs)
        context['show_value'] = self.show_value
        return context



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import logging
from importlib import import_module

from django.apps import AppConfig, apps
from django.conf import settings
from django.db.models import signals

import six

import otree
from otree.models_concrete import UndefinedFormModel, GlobalLockModel
from otree.common_internal import ensure_superuser_exists

logger = logging.getLogger('otree')

# 2017-02-10: move this inside AppConfig.ready?
import_module('otree.checks')   # this made that style check work


def create_singleton_objects(sender, **kwargs):
    for ModelClass in (UndefinedFormModel, GlobalLockModel):
        # if it doesn't already exist, create one.
        ModelClass.objects.get_or_create()


def monkey_patch_static_tag():
    '''
    In Django >= 1.10, you can use {% load static %}
    instead of {% load staticfiles %}. if we switch to that format,
    then it will bypass this. so eventually after Django 1.10, we
    should change this code to patch django.templatetags.static.static
    '''

    from django.contrib.staticfiles.storage import staticfiles_storage
    from django.contrib.staticfiles.templatetags import staticfiles

    def patched_static(path):
        '''same 1-line function,
        just tries to give a friendlier error message'''
        try:
            return staticfiles_storage.url(path)
        except ValueError as exc:
            # Heroku and "otree runprodserver" both execute collectstatic
            # automatically, so there is ordinarily no need to suggest
            # running collectstatic if a file is not found. It's more likely
            # that the file doesn't exist or wasn't added in git.
            if 'runserver' in sys.argv:
                msg = '{} - did you remember to run "otree collectstatic"?'
                raise ValueError(msg.format(exc)) from None
            else:
                raise exc from None


    staticfiles.static = patched_static


def monkey_patch_db_cursor():
    '''Monkey-patch the DB cursor, to catch ProgrammingError and
    OperationalError. The alternative is to use middleware, but (1)
    that doesn't catch errors raised outside of views, like channels consumers
    and the task queue, and (2) it's not as specific, because there are
    OperationalErrors that come from different parts of the app that are
    unrelated to resetdb. This is the most targeted location.
    '''


    # This is actually a method
    # it seems safe to monkey patch, because
    # it hasn't changed for several releases.
    # just a ~5-line function.
    def execute(self, sql, params=None):
        self.db.validate_no_broken_transaction()
        with self.db.wrap_database_errors:
            if params is None:
                return self.cursor.execute(sql)
            else:
                try:
                    return self.cursor.execute(sql, params)
                except Exception as exc:
                    ExceptionClass = type(exc)
                    # it seems there are different exceptions all named
                    # OperationalError (django.db.OperationalError,
                    # sqlite.OperationalError, mysql....)
                    # so, simplest to use the string name
                    if ExceptionClass.__name__ in (
                            'OperationalError', 'ProgrammingError'):
                        # these error messages are localized, so we can't
                        # just check for substring 'column' or 'table'
                        # all the ProgrammingError and OperationalError
                        # instances I've seen so far are related to resetdb
                        tb = sys.exc_info()[2]
                        raise ExceptionClass('{} - try running "otree resetdb".'.format(
                                exc)).with_traceback(tb) from None
                    else:
                        raise

    from django.db.backends import utils
    utils.CursorWrapper.execute = execute


def setup_create_default_superuser():
    authconfig = apps.get_app_config('auth')
    signals.post_migrate.connect(
        ensure_superuser_exists,
        sender=authconfig,
        dispatch_uid='common.models.create_testuser'
    )


def setup_create_singleton_objects():
    signals.post_migrate.connect(create_singleton_objects,
                                 dispatch_uid='create_singletons')


def patch_raven_config():
    # patch settings with info that is only available
    # after other settings loaded
    if hasattr(settings, 'RAVEN_CONFIG'):
        settings.RAVEN_CONFIG['release'] = '{}{}'.format(
            otree.get_version(),
            # need to pass the server if it's DEBUG
            # mode. could do this in extra context or tags,
            # but this seems the most straightforward way
            ',dbg' if settings.DEBUG else ''
        )


class OtreeConfig(AppConfig):
    name = 'otree'
    label = 'otree'
    verbose_name = "oTree"

    def ready(self):
        setup_create_singleton_objects()
        setup_create_default_superuser()
        patch_raven_config()
        monkey_patch_static_tag()
        monkey_patch_db_cursor()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

import hashlib
import logging
import os
import random
import re
import string
import uuid
from collections import OrderedDict
from importlib import import_module
from os.path import dirname, join

import channels
import six
from django.apps import apps
from django.conf import settings
from django.core.urlresolvers import reverse
from django.db import connection
from django.http import HttpResponseRedirect
from django.template.defaultfilters import title
from django.utils.safestring import mark_safe
from six.moves import urllib
from huey.contrib.djhuey import HUEY
import contextlib
from django.db import transaction
# set to False if using runserver
USE_REDIS = True


def add_params_to_url(url, params):
    url_parts = list(urllib.parse.urlparse(url))

    # use OrderedDict because sometimes we want certain params at end
    # for readability/consistency
    query = OrderedDict(urllib.parse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urllib.parse.urlencode(query)
    return urllib.parse.urlunparse(url_parts)


def id_label_name(id, label):
    if label:
        return '{} (label: {})'.format(id, label)
    return '{}'.format(id)


def git_commit_timestamp():
    root_dir = dirname(settings.BASE_DIR)
    try:
        with open(join(root_dir, 'git_commit_timestamp'), 'r') as f:
            return f.read().strip()
    except IOError:
        return ''


SESSION_CODE_CHARSET = string.ascii_lowercase + string.digits


def random_chars(num_chars):
    return ''.join(random.choice(SESSION_CODE_CHARSET) for _ in range(num_chars))


def random_chars_8():
    return random_chars(8)


def random_chars_10():
    return random_chars(10)


def app_name_format(app_name):
    app_label = app_name.split('.')[-1]
    return title(app_label.replace("_", " "))


def directory_name(path):
    return os.path.basename(os.path.normpath(path))


def get_models_module(app_name):
    module_name = '{}.models'.format(app_name)
    return import_module(module_name)


def get_bots_module(app_name):
    try:
        bots_module_name = '{}.bots'.format(app_name)
        bots_module = import_module(bots_module_name)
    except ImportError:
        bots_module_name = '{}.tests'.format(app_name)
        bots_module = import_module(bots_module_name)
    return bots_module


def get_views_module(app_name):
    module_name = '{}.views'.format(app_name)
    return import_module(module_name)


def get_app_constants(app_name):
    '''Return the ``Constants`` object of a app defined in the models.py file.

    Example::

        >>> from otree.common_internal import get_app_constants
        >>> get_app_constants('demo')
        <class demo.models.Constants at 0x7fed46bdb188>

    '''
    return get_models_module(app_name).Constants


def get_dotted_name(Cls):
    return '{}.{}'.format(Cls.__module__, Cls.__name__)


def get_app_label_from_import_path(import_path):
    '''App authors must not override AppConfig.label'''
    return import_path.split('.')[-2]


def get_app_label_from_name(app_name):
    '''App authors must not override AppConfig.label'''
    return app_name.split('.')[-1]


def get_app_name_from_label(app_label):
    '''
    >>> get_app_name_from_label('simple')
    'tests.simple'

    '''
    return apps.get_app_config(app_label).name


def expand_choice_tuples(choices):
    '''allows the programmer to define choices as a list of values rather
    than (value, display_value)

    '''
    if not choices:
        return None
    if not isinstance(choices[0], (list, tuple)):
        choices = [(value, value) for value in choices]
    return choices


def contract_choice_tuples(choices):
    '''Return only values of a choice tuple. If the choices are simple lists
    without display name the same list is returned

    '''
    if not choices:
        return None
    if not isinstance(choices[0], (list, tuple)):
        return choices
    return [value for value, _ in choices]


def min_players_multiple(players_per_group):
    ppg = players_per_group

    if isinstance(ppg, six.integer_types) and ppg >= 1:
        return ppg
    if isinstance(ppg, (list, tuple)):
        return sum(ppg)
    # else, it's probably None
    return 1


def db_table_exists(table_name):
    """Return True if a table already exists"""
    return table_name in connection.introspection.table_names()


def missing_db_table():
    """Try to execute a simple select * for every model registered
    """
    for Model in apps.get_models():
        table_name = Model._meta.db_table
        if not db_table_exists(table_name):
            return '{}.{}'.format(Model._meta.app_label, Model.__name__)


def make_hash(s):
    s += settings.SECRET_KEY
    return hashlib.sha224(s.encode()).hexdigest()[:8]


def get_admin_secret_code():
    s = settings.SECRET_KEY
    return hashlib.sha224(s.encode()).hexdigest()[:8]


def channels_create_session_group_name(pre_create_id):
    return 'wait_for_session_{}'.format(pre_create_id)


def channels_wait_page_group_name(session_pk, page_index,
                                  group_id_in_subsession=''):

    return 'wait-page-{}-page{}-{}'.format(
        session_pk, page_index, group_id_in_subsession)


def channels_group_by_arrival_time_group_name(session_pk, page_index):
    return 'group_by_arrival_time_session{}_page{}'.format(
        session_pk, page_index)


def validate_alphanumeric(identifier, identifier_description):
    if re.match(r'^[a-zA-Z0-9_]+$', identifier):
        return identifier
    raise ValueError(
        '{} "{}" can only contain letters, numbers, '
        'and underscores (_)'.format(
            identifier_description,
            identifier
        )
    )


def create_session_and_redirect(session_kwargs):
    pre_create_id = uuid.uuid4().hex
    session_kwargs['_pre_create_id'] = pre_create_id
    channels_group_name = channels_create_session_group_name(
        pre_create_id)
    channels.Channel('otree.create_session').send({
        'kwargs': session_kwargs,
        'channels_group_name': channels_group_name
    })

    wait_for_session_url = reverse(
        'WaitUntilSessionCreated', args=(pre_create_id,)
    )
    return HttpResponseRedirect(wait_for_session_url)


def ensure_superuser_exists(*args, **kwargs):
    """
    Creates our default superuser, returns True for success
    and False for failure
    """
    from django.contrib.auth.models import User
    username = settings.ADMIN_USERNAME
    password = settings.ADMIN_PASSWORD
    logger = logging.getLogger('otree')
    if User.objects.filter(username=username).exists():
        # msg = 'Default superuser exists.'
        # logger.info(msg)
        return True
    if not password:
        return False
    assert User.objects.create_superuser(username, email='',
                                         password=password)
    msg = 'Created superuser "{}"'.format(username)
    logger.info(msg)
    return True


def release_any_stale_locks():
    '''
    Need to release locks in case the server was stopped abruptly,
    and the 'finally' block in each lock did not execute
    '''
    from otree.models_concrete import GlobalLockModel, ParticipantLockModel
    for LockModel in [GlobalLockModel, ParticipantLockModel]:
        try:
            LockModel.objects.filter(locked=True).update(locked=False)
        except:
            # if server is started before DB is synced,
            # this will raise
            # django.db.utils.OperationalError: no such table:
            # otree_globallockmodel
            # we can ignore that because we just want to make sure there are no
            # active locks
            pass


def get_redis_conn():
    '''reuse Huey Redis connection'''
    return HUEY.storage.conn

def has_group_by_arrival_time(app_name):
    page_sequence = get_views_module(app_name).page_sequence
    if len(page_sequence) == 0:
        return False
    return getattr(page_sequence[0], 'group_by_arrival_time', False)


@contextlib.contextmanager
def transaction_except_for_sqlite():
    '''
    On SQLite, transactions tend to result in "database locked" errors.
    So, skip the transaction on SQLite, to allow local dev.
    Should only be used if omitting the transaction rarely causes problems.
    '''
    if settings.DATABASES['default']['ENGINE'].endswith('sqlite3'):
        yield
    else:
        with transaction.atomic():
            yield


class DebugTable(object):
    def __init__(self, title, rows):
        self.title = title
        self.rows = []
        for k, v in rows:
            if isinstance(v, six.string_types):
                v = v.strip().replace("\n", "<br>")
                v = mark_safe(v)
            self.rows.append((k, v))


# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants


class MyWait(WaitPage):
    group_by_arrival_time = True

    def after_all_players_arrive(self):
        self.group.set_in_wait_page = True
        for p in self.group.get_players():
            p.set_in_wait_page = True


class Page1(Page):

    def before_next_page(self):
        players_in_group = len(self.group.get_players())
        assert players_in_group == Constants.players_per_group, [self.subsession.round_number, self.player.id_in_subsession]


page_sequence = [
    MyWait,
    Page1,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# =============================================================================
# IMPORTS
# =============================================================================

from django.http import HttpResponseServerError

from otree.common_internal import missing_db_table


class CheckDBMiddleware(object):

    synced = None

    def process_request(self, request):
        if not CheckDBMiddleware.synced:
            missing_table = missing_db_table()
            if missing_table:
                print(missing_table, type(missing_table))
                msg = ("Your database is not ready "
                       "(missing table for model {}). "
                       "Try running 'otree resetdb'.").format(missing_table)
                return HttpResponseServerError(msg)
            else:
                CheckDBMiddleware.synced = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import  # for channels module
import os
import channels.asgi

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

from otree.common_internal import (
    release_any_stale_locks, get_redis_conn  # noqa
)
release_any_stale_locks()

# clear any tasks in Huey DB, so they don't pile up over time,
# especially if you run the server without the timeoutworker to consume the
# tasks.
# ideally we would only schedule a task in Huey if timeoutworker is running,
# so that we don't pile up messages that never get consumed, but I don't know
# how and when to check if Huey is running, in a performant way.
# this code is also in timeoutworker.
from huey.contrib.djhuey import HUEY  # noqa
HUEY.flush()

from otree.bots.browser import redis_flush_bots  # noqa
redis_flush_bots(get_redis_conn())

channel_layer = channels.asgi.get_channel_layer()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import io
import glob
import types
import inspect
from importlib import import_module
from functools import wraps

from django.apps import apps
from django.conf import settings
from django.core.checks import register, Error, Warning
from django.template import Template
from django.template import TemplateSyntaxError

from otree.api import BasePlayer, BaseGroup, BaseSubsession, Currency
import otree.views.abstract


class Rules(object):
    """A helper class incapsulating common checks.

    Usage:
        rules = Rules(app_config, errors_list)

        # various rule checks, see below for list of rules
        rules.file_exists('some_file.py')
        ...

        # custom checks
        if <your-condition>:
            rules.push_error('...', id='...')

        # using checks as guards
        if rules.module_exists('tests'):
            tests = rules.get_module('tests') # won't fail
            ...

    """

    common_buffer = {}

    def __init__(self, config, errors, id=None):
        self.config = config
        self.errors = errors
        self.id = id

    def rule(meth):
        '''
        wrapper to return True if the method doesn't return anything.
        and False if it does return something.
        '''
        @wraps(meth)
        def wrapper(self, *args, **kwargs):
            res = meth(self, *args, **kwargs)
            if res:
                self.errors.append(res)
                return False
            else:
                return True
        return wrapper

    def error(self, title, **kwargs):
        kwargs.setdefault('obj', self.config.label)
        kwargs.setdefault('id', self.id)
        return Error(title, **kwargs)

    def push_error(self, title, **kwargs):
        return self.errors.append(self.error(title, **kwargs))

    def warning(self, title, **kwargs):
        kwargs.setdefault('obj', self.config.label)
        kwargs.setdefault('id', self.id)
        return Warning(title, **kwargs)

    def push_warning(self, title, **kwargs):
        return self.errors.append(self.warning(title, **kwargs))

    # Helper meythods

    def get_path(self, name):
        return os.path.join(self.config.path, name)

    def get_rel_path(self, name):
        basepath = os.getcwd()
        return os.path.relpath(name, basepath)

    def get_module(self, name):
        return import_module(self.config.name + '.' + name)

    def get_module_attr(self, module, name):
        if not isinstance(module, types.ModuleType):
            module = self.get_module(module)
        return getattr(module, name)

    def get_template_names(self):
        path = self.get_path('templates')
        template_names = []
        for root, dirs, files in os.walk(path):
            for filename in [f for f in files if f.endswith('.html')]:
                template_names.append(os.path.join(root, filename))
        return template_names

    # Rule methods

    @rule
    def file_exists(self, filename):
        if not os.path.isfile(self.get_path(filename)):
            return self.error('No "%s" file found in game folder' % filename)

    @rule
    def dir_exists(self, filename):
        if not os.path.isdir(self.get_path(filename)):
            msg = 'No "%s" directory found in game folder' % filename
            return self.error(msg)

    @rule
    def model_exists(self, name):
        try:
            self.config.get_model(name)
        except LookupError:
            return self.error('Model "%s" not defined' % name)

    @rule
    def module_exists(self, module):
        try:
            module = self.get_module(module)
        except ImportError as e:
            return self.error('Can\'t import module "%s": %s' % (module, e))

    @rule
    def class_exists(self, module, name):
        module = self.get_module(module)
        cls = getattr(module, name, None)
        if not inspect.isclass(cls):
            msg = 'No class "%s" in module "%s"' % (name, module.__name__)
            return self.error(msg)

    @rule
    def template_has_valid_syntax(self, template_name):
        from otree.checks.templates import has_valid_encoding
        from otree.checks.templates import format_source_snippet

        # Only test files that are valid templates.
        if not has_valid_encoding(template_name):
            return

        try:
            with io.open(template_name, 'r', encoding='utf8') as f:
                Template(f.read())
        except (IOError, OSError):
            pass
        except TemplateSyntaxError as error:
            # The django_template_source attribute will only be available on
            # DEBUG = True.
            if hasattr(error, 'django_template_source'):
                template_source, position = error.django_template_source
                snippet = format_source_snippet(
                    template_source.source,
                    arrow_position=position[0])
                message = (
                    'Template syntax error in {template}\n'
                    '\n'
                    '{snippet}\n'
                    '\n'
                    'Error: {error}'.format(
                        template=template_name,
                        error=error,
                        snippet=snippet))
            else:
                message = (
                    'Template syntax error in {template}\n'
                    'Error: {error}\n'
                    'Set "DEBUG = True" to see more details.'.format(
                        template=template_name, error=error))
            return self.error(message)

    @rule
    def template_has_no_dead_code(self, template_name):
        from otree.checks.templates import get_unreachable_content
        from otree.checks.templates import has_valid_encoding

        # Only test files that are valid templates.
        if not has_valid_encoding(template_name):
            return

        try:
            with io.open(template_name, 'r', encoding='utf8') as f:
                compiled_template = Template(f.read())
        except (IOError, OSError, TemplateSyntaxError):
            # Ignore errors that occured during file-read or compilation.
            return

        def format_content(text):
            text = text.strip()
            lines = text.splitlines()
            lines = ['> {0}'.format(line) for line in lines]
            return '\n'.join(lines)

        contents = get_unreachable_content(compiled_template)
        content_bits = '\n\n'.join(
            format_content(bit)
            for bit in contents)
        if contents:
            return self.error(
                'Template contains the following text outside of a '
                '{% block %}. This text will never be displayed.'
                '\n\n' + content_bits,
                obj=os.path.join(self.config.label,
                                 self.get_rel_path(template_name)))

    @rule
    def template_has_valid_encoding(self, template_name):
        from otree.checks.templates import has_valid_encoding

        if not has_valid_encoding(template_name):
            return self.error(
                'The template {template} is not UTF-8 encoded. '
                'Please configure your text editor to always save files '
                'as UTF-8. Then open the file and save it again.'
                .format(template=self.get_rel_path(template_name)))


def _get_all_configs():
    return [
        app
        for app in apps.get_app_configs()
        if app.name in settings.INSTALLED_OTREE_APPS]


def register_rules(tags=(), id=None, once_per_project=False):
    """Transform a function based on rules, to a something
    django.core.checks.register takes.
    Passes Rules instance as first argument.
    """
    def decorator(func):
        @register(*tags)
        @wraps(func)
        def wrapper(app_configs, **kwargs):
            if once_per_project:
                # some checks should only be run once, not for each app
                app_configs = [apps.get_app_config('otree')]
            else:
                # if app_configs list is given (e.g. otree check app1 app2), run on those
                # if it's None, run on all apps
                # (system check API requires this)
                app_configs = app_configs or _get_all_configs()
            errors = []
            for config in app_configs:
                rules = Rules(config, errors, id=id)
                func(rules, **kwargs)
            return errors
        return wrapper
    return decorator


# Checks
@register_rules(id='otree.E001')
def files(rules, **kwargs):
    rules.file_exists('models.py')
    rules.file_exists('views.py')

    if os.path.isdir(rules.get_path('templates')):
        # check for files in templates, but not in templates/<label>
        misplaced_templates = set(glob.glob(
            os.path.join(rules.get_path('templates'), '*.html')
        ))
        misplaced_templates.discard(rules.config.label)
        if misplaced_templates:
            hint = (
                'Move template files from "{app}/templates/" '
                'to "{app}/templates/{app}" subfolder'.format(
                    app=rules.config.label)
            )
            rules.push_error(
                "Templates files in app's root template directory",
                hint=hint, id='otree.E001'
            )


base_model_attrs = {
    'Player': set(dir(BasePlayer)),
    'Group': set(dir(BaseGroup)),
    'Subsession': set(dir(BaseSubsession)),
}

model_field_substitutes = {
    int: 'IntegerField',
    float: 'FloatField',
    bool: 'BooleanField',
    str: 'CharField',
    Currency: 'CurrencyField'
}


@register_rules(id='otree.E002')
def model_classes(rules, **kwargs):
    rules.model_exists('Subsession')
    rules.model_exists('Group')
    rules.model_exists('Player')

    config = rules.config
    Player = config.get_model('Player')
    Group = config.get_model('Group')
    Subsession = config.get_model('Subsession')

    for Model in [Player, Group, Subsession]:
        for attr in dir(Model):
            if attr not in base_model_attrs[Model.__name__]:
                try:
                    _type = type(getattr(Model, attr))
                except AttributeError:
                    # I got "The 'q_country' attribute can only be accessed
                    # from Player instances."
                    # can just filter/ignore these.
                    pass
                else:
                    if _type in model_field_substitutes.keys():
                        rules.push_warning(
                            '{} has attribute "{}", which is not a model field, '
                            'and will therefore not be saved '
                            'to the database.'.format(
                                Model.__name__, attr
                            ),
                            hint='Consider changing to "{} = models.{}(...)"'.format(attr, model_field_substitutes[_type])
                        )
                    # if people just need an iterable of choices for a model field,
                    # they should use a tuple, not list or dict
                    if _type in {list, dict, set}:
                        warning = (
                            '{ModelName}.{attr} is a {type_name}. '
                            'Modifying it during a session (e.g. appending or setting values) '
                            'will have unpredictable results; '
                            'you should use '
                            'session.vars or participant.vars instead. '
                            'Or, if this {type_name} is read-only, '
                            "then it's recommended to move it outside of this class "
                            '(e.g. put it in Constants).'
                        ).format(ModelName=Model.__name__,
                                 attr=attr,
                                 type_name=_type.__name__)

                        rules.push_warning(warning)


@register_rules(id='otree.E003')
def constants(rules, **kwargs):
    cond = (
        rules.module_exists('models') and
        rules.class_exists('models', 'Constants')
    )
    if cond:
        Constants = rules.get_module_attr('models', 'Constants')
        attrs = ['name_in_url', 'players_per_group', 'num_rounds']
        for attr_name in attrs:
            if not hasattr(Constants, attr_name):
                msg = "models.py: 'Constants' class needs to define '{}'"
                rules.push_error(msg.format(attr_name))
        ppg = Constants.players_per_group
        if ppg == 0 or ppg == 1:
            rules.push_error(
                "models.py: 'Constants.players_per_group' cannot be {}. You "
                "should set it to None, which makes the group "
                "all players in the subsession.".format(ppg)
            )


@register_rules(id='otree.E004')
def pages_function(rules, **kwargs):
    if rules.module_exists('views'):
        views_module = rules.get_module('views')
        try:
            page_list = views_module.page_sequence
        except:
            rules.push_error('views.py is missing the variable page_sequence.')
            return
        else:
            for i, ViewCls in enumerate(page_list):
                # there is no good reason to include Page in page_sequence.
                # however, WaitPage could belong there. it works fine currently
                # and can save the effort of subclassing
                if ViewCls.__name__ == 'Page':
                    msg = (
                        "views.py: page_sequence cannot contain "
                        "a class called 'Page'. You should subclass Page "
                        "and give your page a different name."
                    )
                    rules.push_error(msg)
                if issubclass(ViewCls,
                              otree.views.abstract.InGameWaitPageMixin):
                    if hasattr(ViewCls, 'before_next_page'):
                        rules.push_error(
                            'views.py: "{}" defines before_next_page, '
                            'which is not valid on wait pages.'.format(
                                ViewCls.__name__)
                        )
                    if ViewCls.group_by_arrival_time:
                        if i > 0:
                            rules.push_error(
                                'views.py: "{}" has group_by_arrival_time=True, so '
                                'it must be placed first in page_sequence.'.format(
                                    ViewCls.__name__))
                        if ViewCls.wait_for_all_groups:
                            rules.push_error(
                                'views.py: "{}" has group_by_arrival_time=True, so '
                                'it cannot have wait_for_all_groups=True also.'.format(
                                    ViewCls.__name__))

                elif issubclass(ViewCls,
                                otree.views.abstract.FormPageMixin):
                    # ok
                    pass
                else:
                    msg = 'views.py: "{}" is not a valid page'.format(ViewCls)
                    rules.push_error(msg)


@register_rules(id='otree.E005')
def templates_have_no_dead_code(rules, **kwargs):
    for template_name in rules.get_template_names():
        rules.template_has_no_dead_code(template_name)


@register_rules(id='otree.E006', once_per_project=True)
def unique_sessions_names(rules, **kwargs):
    already_seen = set()
    for st in settings.SESSION_CONFIGS:
        st_name = st["name"]
        if st_name in already_seen:
            msg = "Duplicate SESSION_CONFIG name '{}'".format(st_name)
            rules.push_error(msg)
        else:
            already_seen.add(st_name)


@register_rules(id='otree.E009', once_per_project=True)
def unique_room_names(rules, **kwargs):
    already_seen = set()
    for room in getattr(settings, 'ROOMS', []):
        room_name = room["name"]
        if room_name in already_seen:
            msg = "Duplicate ROOM name '{}'".format(room_name)
            rules.push_error(msg)
        else:
            already_seen.add(room_name)


@register_rules(id='otree.E007')
def template_encoding(rules, **kwargs):
    for template_name in rules.get_template_names():
        rules.template_has_valid_encoding(template_name)


@register_rules(id='otree.E008')
def templates_have_valid_syntax(rules, **kwargs):
    for template_name in rules.get_template_names():
        rules.template_has_valid_syntax(template_name)


# TODO: startapp should pass validation checks



from otree.api import Submission, SubmissionMustFail
from .base import TestCase
import otree.session
from otree.bots.runner import session_bot_runner_factory, test_bots
import logging
import unittest.mock
from tests.bots_cases.tests import PlayerBot
import tests.bots_cases.views

class TestBots(TestCase):
    def test_bot_runs(self):

        session = otree.session.create_session(
            session_config_name='bots_raise',
            num_participants=2,
            use_cli_bots=True,
            bot_case_number=0,
        )

        bot_runner = session_bot_runner_factory(session)

        with self.assertRaises(ZeroDivisionError):
            bot_runner.play()

    def test_bots_check_html(self):

        session = otree.session.create_session(
            session_config_name='bots_check_html',
            num_participants=1,
            use_cli_bots=True,
            bot_case_number=0,
        )

        with self.settings(BOTS_CHECK_HTML=True):
            from django.conf import settings
            self.assertEqual(settings.BOTS_CHECK_HTML, True)

            bot_runner = session_bot_runner_factory(session)

            try:
                bot_runner.play()
            except AssertionError as exc:
                # AssertionError should say something about check_html
                raises_correct_message = 'check_html' in str(exc)
            else:
                raises_correct_message = False
            if not raises_correct_message:
                raise AssertionError('bots check_html not working properly')

    def test_bot_bad_post(self):
        """
        Test that posting bad data without using SubmitInvalid
        will raise an error.
        """

        session = otree.session.create_session(
            session_config_name='bots_bad_post',
            num_participants=1,
            use_cli_bots=True,
            bot_case_number=0,
        )

        bot_runner = session_bot_runner_factory(session)

        with self.assertRaises(AssertionError):
            # need to disable log output, because this triggers an exception
            # that is logged to stdout
            logging.disable(logging.CRITICAL)
            bot_runner.play()
            logging.disable(logging.NOTSET)

    def test_bots_submission_varieties(self):
        '''Testing different syntaxes for a submission'''
        Page1 = tests.bots_cases.views.Page1
        for submission in [
            Submission(Page1),
            Submission(Page1, {}),
            SubmissionMustFail(Page1, check_html=False),
            Submission(Page1, {'f2': True}, check_html=False),
            SubmissionMustFail(Page1, {}),
            Submission(Page1, {'f3': True}),
        ]:
            self.assertIsInstance(submission, dict)

    @unittest.mock.patch.object(PlayerBot, 'case1')
    @unittest.mock.patch.object(PlayerBot, 'case2')
    def test_cases(self, patched_case2, patched_case1):
        '''
        Test that all cases are run
        '''

        test_bots('bots_cases', 1, False)
        self.assertTrue(patched_case1.called)
        self.assertTrue(patched_case2.called)



#!/usr/bin/env python
# encoding: utf-8

from otree.export import get_field_names_for_live_update
from rest_framework import serializers
from otree.models import Participant


class ParticipantSerializer(serializers.ModelSerializer):
    class Meta:
        model = Participant
        fields = get_field_names_for_live_update(Participant)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import otree.db.models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0002_browserbotsubmit_is_last'),
    ]

    operations = [
        migrations.AddField(
            model_name='participant',
            name='_browser_bot_finished',
            field=otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')]),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import decimal
import logging
import abc
import six
from six.moves import urllib
from six.moves.html_parser import HTMLParser

from django import test
from django.core.urlresolvers import resolve
from django.conf import settings
from easymoney import Money as Currency

from otree import constants_internal

from otree.common_internal import (
    get_dotted_name, get_bots_module, get_admin_secret_code)

ADMIN_SECRET_CODE = get_admin_secret_code()

logger = logging.getLogger('otree.bots')

DISABLE_CHECK_HTML_INSTRUCTIONS = '''
Checking the HTML may not find all form fields and buttons
(e.g. those added with JavaScript),
so you can disable this check by yielding a Submission
with check_html=False, e.g.:

yield Submission(views.PageName, {{...}}, check_html=False)
'''

HTML_MISSING_BUTTON_WARNING = ('''
Bot is trying to submit page {page_name},
but no button was not found in the HTML of the page.
(searched for <input> with type='submit' or <button> with type != 'button').
''' + DISABLE_CHECK_HTML_INSTRUCTIONS).replace('\n', ' ').strip()

HTML_MISSING_FIELD_WARNING = ('''
Bot is trying to submit page {page_name} with fields: "{fields}",
but these form fields were not found in the HTML of the page
(searched for tags {tags} with name= attribute matching the field name).
''' + DISABLE_CHECK_HTML_INSTRUCTIONS).replace('\n', ' ').strip()


class BOTS_CHECK_HTML:
    pass


def _Submission(
        PageClass, post_data=None, *, check_html=BOTS_CHECK_HTML,
        must_fail=False, timeout_happened=False):

    post_data = post_data or {}

    # don't mutate the input
    post_data = post_data.copy()

    if check_html == BOTS_CHECK_HTML:
        check_html = settings.BOTS_CHECK_HTML

    if must_fail:
        # must_fail needs to go in post_data rather than being a separate
        # dict key, because CLI bots and browser bots need to work the same way.
        # CLI bots can only talk to server through post data
        post_data['must_fail'] = True

    if timeout_happened:
        post_data[constants_internal.timeout_happened] = True
        post_data[constants_internal.admin_secret_code] = ADMIN_SECRET_CODE

    # easy way to check if it's a wait page, without any messy imports
    if hasattr(PageClass, 'wait_for_all_groups'):
        raise AssertionError(
            "Your bot yielded '{}', which is a wait page. "
            "You should delete this line, because bots handle wait pages "
            "automatically.".format(PageClass)
        )

    for key in post_data:
        if isinstance(post_data[key], Currency):
            # because must be json serializable for Huey
            post_data[key] = str(decimal.Decimal(post_data[key]))

    return {
        'page_class': PageClass,
        'page_class_dotted': get_dotted_name(PageClass),
        'post_data': post_data,
        'check_html': check_html
    }


def Submission(
        PageClass, post_data=None, *, check_html=BOTS_CHECK_HTML,
        timeout_happened=False):
    return _Submission(
        PageClass, post_data, check_html=check_html,
        timeout_happened=timeout_happened)


def SubmissionMustFail(
        PageClass, post_data=None, *, check_html=BOTS_CHECK_HTML):
    '''lets you intentionally submit with invalid
    input to ensure it's correctly rejected'''

    return _Submission(
        PageClass,
        post_data=post_data, check_html=check_html, must_fail=True)


def BareYieldToSubmission(yielded_value):

    post_data = {}

    if isinstance(yielded_value, (list, tuple)):
        PageClass = yielded_value[0]
        if len(yielded_value) == 2:
            # shouldn't mutate the input
            post_data = yielded_value[1]
    else:
        PageClass = yielded_value

    return Submission(PageClass, post_data)


def normalize_html_whitespace(html):
    html = html.replace('\n', ' ').replace('\r', ' ')
    html = re.sub(r'\s+', ' ', html)
    return html


class HtmlString(str):

    def truncated(self):
        '''
        Make output more readable by truncating everything before the
         {% content %} block. I also considered indenting the HTML,
         but minidom had a parse error, and BS4 modifies a lot of tags,
         didn't seem optimal.
        '''
        div_str = '<div id="otree-content">'
        i = self.index(div_str) + len(div_str)
        return '...' + self[i:]

    def __str__(self):
        return self.truncated()

    def __repr__(self):
        return self.truncated()


# inherit from object for Python2.7 support.
# otherwise, get
class PageHtmlChecker(HTMLParser, object):

    def __init__(self, fields_to_check):
        super(PageHtmlChecker, self).__init__()
        self.missing_fields = set(fields_to_check)
        self.field_tags = {'input', 'button', 'select', 'textarea'}
        self.submit_button_found = False

    def get_missing_fields(self, html):
        self.feed(html)
        return self.missing_fields

    def check_if_field(self, tag, attrs):
        if tag in self.field_tags:
            for (attr_name, attr_value) in attrs:
                if attr_name == 'name':
                    self.missing_fields.discard(attr_value)

    def check_if_button(self, tag, attrs):
        if not self.submit_button_found:
            if tag == 'button':
                for (attr_name, attr_value) in attrs:
                    if attr_name == 'type' and attr_value == 'button':
                        return
                self.submit_button_found = True
            if tag == 'input':
                for (attr_name, attr_value) in attrs:
                    if attr_name == 'type' and attr_value == 'submit':
                        self.submit_button_found = True

    def handle_starttag(self, tag, attrs):
        self.check_if_field(tag, attrs)
        self.check_if_button(tag, attrs)


def is_wait_page(response):
    return (
        response.get(constants_internal.wait_page_http_header) ==
        constants_internal.get_param_truth_value)


def refresh_from_db(obj):
    return type(obj).objects.get(pk=obj.pk)


class ParticipantBot(six.with_metaclass(abc.ABCMeta, test.Client)):

    def __init__(
            self, participant, load_player_bots=True, **kwargs):
        self.participant = participant
        self.url = None
        self._response = None
        self._html = None
        self.path = None
        self.submits = None
        super(ParticipantBot, self).__init__()

        self.player_bots = []

        # load_player_bots can be set to False when it's convenient for
        # internal testing
        if load_player_bots:
            for player in self.participant.get_players():
                bots_module = get_bots_module(player._meta.app_config.name)
                player_bot = bots_module.PlayerBot(
                    player=player,
                    participant_bot=self)
                self.player_bots.append(player_bot)
            self.submits_generator = self.get_submits()

    def open_start_url(self):
        self.response = self.get(
            self.participant._start_url(),
            follow=True
        )

    def get_submits(self):
        for player_bot in self.player_bots:
            # play_round populates legacy submit list
            generator = player_bot.play_round()
            if player_bot._legacy_submit_list:
                for submission in player_bot._legacy_submit_list:
                    yield submission
            else:
                try:
                    for submission in generator:
                        # Submission or SubmissionMustFail returns a dict
                        # so, we normalize to a dict
                        if not isinstance(submission, dict):
                            submission = BareYieldToSubmission(submission)
                        self.assert_correct_page(submission)
                        self.assert_html_ok(submission)
                        yield submission
                # handle the case where it's empty
                except TypeError as exc:
                    if 'is not iterable' in str(exc):
                        # we used to raise StopIteration here. But shouldn't
                        # do that, because then the whole participant bot
                        # stops running (e.g. doesn't play any of the
                        # PlayerBots in the following apps).
                        # this was causing a bug where we got "bot completed"
                        # but the bot had only played half the game
                        pass
                    else:
                        raise

    def assert_html_ok(self, submission):
        if submission['check_html']:
            field_names = [
                f for f in submission['post_data'].keys() if f != 'must_fail']
            checker = PageHtmlChecker(field_names)
            missing_fields = checker.get_missing_fields(self.html)
            if missing_fields:
                page_name = submission['page_class'].url_name()
                raise AssertionError(
                    HTML_MISSING_FIELD_WARNING.format(
                        page_name=page_name,
                        fields=', '.join(missing_fields),
                        tags=', '.join('<{}>'.format(tag)
                                       for tag in checker.field_tags)))
            if not checker.submit_button_found:
                page_name = submission['page_class'].url_name()
                raise AssertionError(HTML_MISSING_BUTTON_WARNING.format(
                    page_name=page_name))

    def assert_correct_page(self, submission):
        PageClass = submission['page_class']
        expected_url = PageClass.url_name()
        actual_url = resolve(self.path).url_name

        if not expected_url == actual_url:
            raise AssertionError(
                "Bot expects to be on page {}, "
                "but current page is {}. "
                "Check your bot in tests.py, "
                "then create a new session.".format(expected_url, actual_url))

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        try:
            self.url = response.redirect_chain[-1][0]
        except IndexError as exc:
            # this happens e.g. if you use SubmissionMustFail
            # and it returns the same URL
            pass
        else:
            self.path = urllib.parse.urlsplit(self.url).path
        self._response = response
        self.html = response.content.decode('utf-8')

    @property
    def html(self):
        return self._html

    @html.setter
    def html(self, html):
        self._html = HtmlString(normalize_html_whitespace(html))

    def on_wait_page(self):
        # if the existing response was a form page, it will still be...
        # no need to check again
        if not is_wait_page(self.response):
            return False

        # however, wait pages can turn into regular pages, so let's try again
        self.response = self.get(self.url, follow=True)
        return is_wait_page(self.response)

    def submit(self, submission):
        post_data = submission['post_data']
        pretty_post_data = bot_prettify_post_data(post_data)
        log_string = self.path
        if pretty_post_data:
            log_string += ', {}'.format(pretty_post_data)
        if post_data.get('must_fail'):
            log_string += ', SubmissionMustFail'
        if post_data.get('timeout_happened'):
            log_string += ', timeout_happened'
        logger.info(log_string)
        self.response = self.post(self.url, post_data, follow=True)


class PlayerBot(object):

    cases = []

    def __init__(self, player, participant_bot, **kwargs):

        self.participant_bot = participant_bot
        self._cached_player = player
        self._cached_subsession = player.subsession
        self._cached_participant = player.participant
        self._cached_session = player.session
        self._legacy_submit_list = []

        case_number = self._cached_session._bot_case_number
        cases = self.cases
        if len(cases) >= 1:
            self.case = cases[case_number % len(cases)]
        else:
            self.case = None

    def play_round(self):
        pass

    @property
    def player(self):
        return refresh_from_db(self._cached_player)

    @property
    def group(self):
        # the group can change, so can't use cached version
        return self.player.group

    @property
    def subsession(self):
        return refresh_from_db(self._cached_subsession)

    @property
    def session(self):
        return refresh_from_db(self._cached_session)

    @property
    def participant(self):
        return refresh_from_db(self._cached_participant)

    def submit(self, ViewClass, param_dict=None):
        self._legacy_submit_list.append(
            _Submission(ViewClass, param_dict, check_html=False))

    def submit_invalid(self, ViewClass, param_dict=None):
        # simpler to make this a no-op, it makes porting to yield easier
        # then we can just do a search-and-replace
        # self._legacy_submit_list.append((ViewClass, param_dict, 'invalid'))
        pass

    @property
    def html(self):
        return self.participant_bot.html


def bot_prettify_post_data(post_data):
    if hasattr(post_data, 'dict'):
        # if using CLI bots, this will be a
        # MultiValueKeyDict, because that's what request.POST
        # contains. we need to turn it into a regular dict
        # (i.e. values should not be single-element lists)
        post_data = post_data.dict()

    post_data = post_data.copy()
    for extra_key in [
        'csrfmiddlewaretoken', 'origin_url', 'must_fail', 'timeout_happened',
        'admin_secret_code'
    ]:
        post_data.pop(extra_key, None)

    # if browser bots, it will be a regular dict
    return post_data



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants


class GBATWaitPage(WaitPage):
    group_by_arrival_time = True

    def get_players_for_group(self, waiting_players):
        for attr in ['player', 'participant', 'group']:
            try:
                getattr(self, attr).id
            except AttributeError:
                pass
            else:
                raise AssertionError(
                    'You should not be able to reference '
                    'self.{player, participant, group} '
                    'in get_players_for_group'
                )
        if len(waiting_players) == 2:
            return waiting_players

class RegularWait(WaitPage):
    def after_all_players_arrive(self):
        for attr in ['player', 'participant']:
            try:
                obj = getattr(self, attr)
                obj.id = 100
            except AttributeError:
                pass
            else:
                raise AssertionError(
                    'You should not be able to reference self.participant '
                    'or self.player in after_all_players_arrive'
                )


class AllGroupsWait(WaitPage):
    wait_for_all_groups = True

    def after_all_players_arrive(self):
        for attr in ['player', 'participant', 'group']:
            try:
                obj = getattr(self, attr)
                obj.id = 100
            except AttributeError:
                pass
            else:
                raise AssertionError(
                    'You should not be able to reference '
                    'self.{player, participant, group} '
                    'in this method'
                )



class Page1(Page):
    pass


page_sequence = [
    GBATWaitPage,
    RegularWait,
    AllGroupsWait,
    Page1,
]



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import

from otree.api import WaitPage

from tests.utils import BlankTemplatePage as Page

from .models import Constants


class MyPage(Page):
    def is_displayed(self):
        return self.round_number == Constants.num_rounds


class ResultsWaitPage(WaitPage):
    def is_displayed(self):
        return self.round_number == Constants.num_rounds


class Results(Page):
    def is_displayed(self):
        return self.round_number == Constants.num_rounds


page_sequence = [
    MyPage,
    ResultsWaitPage,
    Results
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path

from django.conf import global_settings
from django.contrib.messages import constants as messages

from six.moves.urllib import parse as urlparse


DEFAULT_MIDDLEWARE_CLASSES = (
    'otree.middleware.CheckDBMiddleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    # this middlewware is for generate human redeable errors

    'django.contrib.sessions.middleware.SessionMiddleware',
    # 'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',

    # 2015-04-08: disabling SSLify until we make this work better
    # 'sslify.middleware.SSLifyMiddleware',
)


def collapse_to_unique_list(*args):
    """Create a new list with all elements from a given lists without reapeated
    elements

    """
    combined = []
    for arg in args:
        for elem in arg or ():
            if elem not in combined:
                combined.append(elem)
    return combined


def get_default_settings(initial_settings=None):
    if initial_settings is None:
        initial_settings = {}
    logging = {
        'version': 1,
        'disable_existing_loggers': False,
        'root': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'formatters': {
            'verbose': {
                'format': '[%(levelname)s|%(asctime)s] %(name)s > %(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'verbose'
            },
            'sentry': {
                'level': 'WARNING',
                'class': (
                    'raven.contrib.django.raven_compat.handlers.'
                    'SentryHandler'
                ),
            },
        },
        'loggers': {
            'otree.test.core': {
                'handlers': ['console'],
                'propagate': False,
                'level': 'INFO',
            },
            # 2016-07-25: botworker seems to be sending messages to Sentry
            # without any special configuration, not sure why.
            # but, i should use a logger, because i need to catch exceptions
            # in botworker so it keeps running
            'otree.test.browser_bots': {
                'handlers': ['sentry', 'console'],
                'propagate': False,
                'level': 'INFO',
            },
            'django.request': {
                'handlers': ['console'],
                'propagate': True,
                'level': 'DEBUG',
            },
            # logger so that we can explicitly send certain warnings to sentry,
            # without raising an exception.
            # 2016-10-23: has not been used yet
            'otree.sentry': {
                'handlers': ['sentry'],
                'propagate': True,
                'level': 'DEBUG',
            },
            # log any error that occurs inside channels code
            'django.channels': {
                'handlers': ['sentry'],
                'propagate': True,
                'level': 'ERROR',
            },
            # This is required for exceptions inside Huey tasks to get logged
            # to Sentry
            'huey.consumer': {
                'handlers': ['sentry', 'console'],
                'level': 'INFO'
            },
            # suppress the INFO message: 'raven is not configured (logging
            # disabled).....', in case someone doesn't have a DSN
            'raven.contrib.django.client.DjangoClient': {
                'handlers': ['console'],
                'level': 'WARNING'
            }


        }
    }

    REDIS_URL = os.environ.get('REDIS_URL', 'redis://localhost:6379')

    return {
        # set to True so that if there is an error in an {% include %}'d
        # template, it doesn't just fail silently. instead should raise
        # an error (and send through Sentry etc)
        'STATIC_ROOT': os.path.join(
            initial_settings.get('BASE_DIR', ''),
            '_static_root'),
        'STATIC_URL': '/static/',
        'STATICFILES_STORAGE': (
            'whitenoise.django.GzipManifestStaticFilesStorage'
        ),
        'ROOT_URLCONF': 'otree.urls',

        'TIME_ZONE': 'UTC',
        'USE_TZ': True,
        'ALLOWED_HOSTS': ['*'],

        'LOGGING': logging,

        'REAL_WORLD_CURRENCY_CODE': 'USD',
        'REAL_WORLD_CURRENCY_LOCALE': 'en_US',
        'REAL_WORLD_CURRENCY_DECIMAL_PLACES': 2,
        'USE_POINTS': True,

        'POINTS_DECIMAL_PLACES': 0,

        # eventually can remove this,
        # when it's present in otree-library
        # that most people downloaded
        'USE_L10N': True,
        'SECURE_PROXY_SSL_HEADER': ('HTTP_X_FORWARDED_PROTO', 'https'),
        'MTURK_HOST': 'mechanicalturk.amazonaws.com',
        'MTURK_SANDBOX_HOST': 'mechanicalturk.sandbox.amazonaws.com',

        # The project can override the routing.py used as entry point by
        # setting CHANNEL_ROUTING.

        'CHANNEL_LAYERS': {
            'default': {
                "BACKEND": "otree.channels.asgi_redis.RedisChannelLayer",
                "CONFIG": {
                    "hosts": [REDIS_URL],
                },
                'ROUTING': initial_settings.get(
                    'CHANNEL_ROUTING',
                    'otree.channels.routing.channel_routing'),
            },
            'inmemory': {
                "BACKEND": "asgiref.inmemory.ChannelLayer",
                'ROUTING': initial_settings.get(
                    'CHANNEL_ROUTING',
                    'otree.channels.routing.channel_routing'),
            },
        },

        # for convenience within oTree
        'REDIS_URL': REDIS_URL,

        # since workers on Amazon MTurk can return the hit
        # we need extra participants created on the
        # server.
        # The following setting is ratio:
        # num_participants_server / num_participants_mturk
        'MTURK_NUM_PARTICIPANTS_MULTIPLE': 2,
        'LOCALE_PATHS': [
            os.path.join(initial_settings.get('BASE_DIR', ''), 'locale')
        ],

        # ideally this would be a per-app setting, but I don't want to
        # pollute Constants. It doesn't make as much sense per session config,
        # so I'm just going the simple route and making it a global setting.
        'BOTS_CHECK_HTML': True,
    }


def augment_settings(settings):

    if 'POINTS_CUSTOM_NAME' in settings:
        settings.setdefault(
            'POINTS_CUSTOM_FORMAT',
            '{} ' + settings['POINTS_CUSTOM_NAME']
        )

    all_otree_apps_set = set()

    if ('SESSION_CONFIGS' not in settings and
            'SESSION_TYPES' in settings):
        raise ValueError(
            'In settings.py, you should rename '
            'SESSION_TYPES to SESSION_CONFIGS, and '
            'SESSION_TYPE_DEFAULTS to SESSION_CONFIG_DEFAULTS.'
        )

    for s in settings['SESSION_CONFIGS']:
        for app in s['app_sequence']:
            all_otree_apps_set.add(app)

    all_otree_apps = list(all_otree_apps_set)

    no_experiment_apps = [
        'django.contrib.auth',
        'otree',
        'floppyforms',
        # need this for admin login
        'django.contrib.admin',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.messages',
        'django.contrib.staticfiles',
        'otree.timeout',
        'channels',
        'huey.contrib.djhuey',
        'rest_framework',
        'idmap',
    ]

    if settings.get('SENTRY_DSN'):
        settings.setdefault(
            'RAVEN_CONFIG',
            {
                'dsn': settings['SENTRY_DSN'],
                'processors': ['raven.processors.SanitizePasswordsProcessor'],
            }
        )
        no_experiment_apps.append('raven.contrib.django.raven_compat')

    # order is important:
    # otree unregisters User & Group, which are installed by auth.
    # otree templates need to get loaded before the admin.
    no_experiment_apps = collapse_to_unique_list(
        no_experiment_apps,
        settings['INSTALLED_APPS']
    )

    new_installed_apps = collapse_to_unique_list(
        no_experiment_apps, all_otree_apps)

    additional_template_dirs = []
    template_dir = os.path.join(settings['BASE_DIR'], 'templates')
    # 2015-09-21: i won't put a deprecation warning because 'templates/'
    # is the django convention and someone might legitimately want it.
    # just remove this code at some point
    # same for static/ dir below
    if os.path.exists(template_dir):
        additional_template_dirs = [template_dir]

    _template_dir = os.path.join(settings['BASE_DIR'], '_templates')
    if os.path.exists(_template_dir):
        additional_template_dirs = [_template_dir]

    new_template_dirs = collapse_to_unique_list(
        settings.get('TEMPLATE_DIRS'),
        # 2015-5-2: 'templates' is deprecated in favor of '_templates'
        # remove it at some point
        additional_template_dirs,
    )

    static_dir = os.path.join(settings['BASE_DIR'], 'static')
    additional_static_dirs = []
    if os.path.exists(static_dir):
        additional_static_dirs = [static_dir]

    _static_dir = os.path.join(settings['BASE_DIR'], '_static')
    if os.path.exists(_static_dir):
        additional_static_dirs = [_static_dir]

    new_staticfiles_dirs = collapse_to_unique_list(
        settings.get('STATICFILES_DIRS'),
        # 2015-5-2: 'static' is deprecated in favor of '_static'
        # remove it at some point. or keep it, because that's the Django
        # convention
        additional_static_dirs,
    )

    new_middleware_classes = collapse_to_unique_list(
        DEFAULT_MIDDLEWARE_CLASSES,
        settings.get('MIDDLEWARE_CLASSES'))

    augmented_settings = {
        'INSTALLED_APPS': new_installed_apps,
        'TEMPLATES': [{
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': new_template_dirs,
            'OPTIONS': {
                # 2016-10-08: setting template debug back to True,
                # because if an included template has an error, we need
                # to surface the error, rather than not showing the template.
                # that's how I set it in d1cd00ebfd43c7eff408dea6363fd14bb90e7c06,
                # but then in 2c10188b33f2ac36c046f4f0f8764e15d6a6fa81,
                # i set this to False, but I'm not sure why and there is no
                # note in the commit explaining why.
                'debug': True,
                'loaders': [
                    ('django.template.loaders.cached.Loader', [
                        'django.template.loaders.filesystem.Loader',
                        'django.template.loaders.app_directories.Loader',
                    ]),
                ],
                'context_processors': collapse_to_unique_list(
                    global_settings.TEMPLATE_CONTEXT_PROCESSORS, (
                        'django.core.context_processors.request',
                    )
                ),
            },
        }],
        'STATICFILES_DIRS': new_staticfiles_dirs,
        'MIDDLEWARE_CLASSES': new_middleware_classes,
        'NO_EXPERIMENT_APPS': no_experiment_apps,
        'INSTALLED_OTREE_APPS': all_otree_apps,
        'MESSAGE_TAGS': {messages.ERROR: 'danger'},
        'LOGIN_REDIRECT_URL': 'Sessions',
    }

    settings.setdefault('LANGUAGE_CODE', global_settings.LANGUAGE_CODE)

    CURRENCY_LOCALE = settings.get('REAL_WORLD_CURRENCY_LOCALE', None)
    if not CURRENCY_LOCALE:

        # favor en_GB currency formatting since it represents negative amounts
        # with minus signs rather than parentheses
        # if settings['LANGUAGE_CODE'][:2] == 'en':
        #     CURRENCY_LOCALE = 'en_GB'
        # else:
        CURRENCY_LOCALE = settings['LANGUAGE_CODE']

        settings.setdefault('REAL_WORLD_CURRENCY_LOCALE',
                            CURRENCY_LOCALE.replace('-', '_'))

    overridable_settings = get_default_settings(settings)

    settings.update(augmented_settings)

    for k, v in overridable_settings.items():
        settings.setdefault(k, v)

    redis_url = urlparse.urlparse(settings.get('REDIS_URL'))

    settings['HUEY'] = {
        'name': 'otree-huey',
        'connection': {
            'host': redis_url.hostname,
            'port': redis_url.port,
            'password': redis_url.password
        },
        'always_eager': False,
        # I need a result store to retrieve the results of browser-bots
        # tasks and pinging, even if the result is evaluated immediately
        # (otherwise, calling the task returns None.
        'result_store': False,
        'consumer': {
            'workers': 1,
            # 'worker_type': 'thread',
            'scheduler_interval': 5,
            'loglevel': 'warning',
        },
    }



from django.apps import apps
from importlib import import_module

"""
(THIS IS CURRENTLY PRIVATE API, MAY CHANGE WITHOUT NOTICE)

To create an oTree extension, add a package called ``otree_extensions``
to your app. It can contain any of the following submodules:

urls.py
-------

should contain a variable ``urlpatterns``, which will be appended to
oTree's built-in URL patterns.

routing.py
----------

Should contain a variable ``channel_routing``,
with a list of channel routes, as described in the Django channels documentation:

https://channels.readthedocs.io/en/stable/getting-started.html#routing

These routes will be appended to oTree's built-in channel routes.

admin.py
--------

This module allows you to define custom data exports that will be included
in oTree's data export page. Define a variable ``data_export_views``,
which is a list of Django class-based views (see Django docs).

Each view should define a ``get()`` method with the following signature::

    def get(self, request, *args, **kwargs):

This method should return an HTTP response with
the exported data (e.g. CSV, XLSX, JSON, etc), using the appropriate MIME type
on the HTTP response.

Each view must also have the following attributes:

-   ``url_pattern``: the URL pattern string, e.g. '^mychat_export/$'
-   ``url_name``: see Django docs on reverse resolution of URLs, e.g. 'mychat_export'
-   ``display_name``: The text of the download hyperlink on the data export page
    (e.g. "Chat Data Export")

You don't need to worry about login_required and AUTH_LEVEL;
oTree will handle this automatically.

(In the future, admin.py may be used for other admin customizations,
not just data export.)

"""

def get_extensions_modules(submodule_name):
    modules = []
    for app_config in apps.get_app_configs():
        try:
            dotted_path = '{}.otree_extensions.{}'.format(
                app_config.name, submodule_name)
            module = import_module(dotted_path)
            modules.append(module)
        # ModuleNotFoundError is new in Python 3.6
        except ImportError:
            continue
    return modules

def get_extensions_data_export_views():
    view_classes = []
    for module in get_extensions_modules('admin'):
        view_classes += getattr(module, 'data_export_views', [])
    return view_classes


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import otree.models.varsmixin
import otree.db.models
import otree.common_internal


class Migration(migrations.Migration):

    replaces = [('otree', '0001_initial'), ('otree', '0002_browserbotsubmit_is_last'), ('otree', '0003_participant__browser_bot_finished'), ('otree', '0004_auto_20160704_1636'), ('otree', '0005_auto_20160707_1914'), ('otree', '0006_auto_20160708_0657'), ('otree', '0007_auto_20160726_1956'), ('otree', '0008_auto_20160728_1916'), ('otree', '0009_browserbotslaunchersessioncode'), ('otree', '0010_session__bot_case_number'), ('otree', '0011_auto_20160805_1735'), ('otree', '0012_auto_20160821_2339'), ('otree', '0013_failedsessioncreation_traceback'), ('otree', '0014_auto_20161012_0539'), ('otree', '0015_auto_20161111_1055'), ('otree', '0016_auto_20161120_1843'), ('otree', '0017_auto_20170208_0103')]

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CompletedGroupWaitPage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('page_index', models.PositiveIntegerField()),
                ('group_pk', models.PositiveIntegerField()),
                ('after_all_players_arrive_run', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='CompletedSubsessionWaitPage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('page_index', models.PositiveIntegerField()),
                ('after_all_players_arrive_run', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='ExpectedRoomParticipant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('room_name', models.CharField(max_length=50)),
                ('participant_label', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='FailedSessionCreation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('pre_create_id', models.CharField(max_length=100, db_index=True)),
                ('message', models.CharField(max_length=300)),
                ('traceback', models.TextField(default='')),
            ],
        ),
        migrations.CreateModel(
            name='PageCompletion',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('app_name', models.CharField(max_length=300)),
                ('page_index', models.PositiveIntegerField()),
                ('page_name', models.CharField(max_length=300)),
                ('time_stamp', models.PositiveIntegerField()),
                ('seconds_on_page', models.PositiveIntegerField()),
                ('subsession_pk', models.PositiveIntegerField()),
                ('auto_submitted', models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name='PageTimeout',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('page_index', models.PositiveIntegerField()),
                ('expiration_time', models.PositiveIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Participant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('vars', otree.db.models._JSONField(null=True, default=dict)),
                ('exclude_from_data_analysis', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('time_started', otree.db.models.DateTimeField(null=True)),
                ('mturk_assignment_id', otree.db.models.CharField(max_length=50, null=True)),
                ('mturk_worker_id', otree.db.models.CharField(max_length=50, null=True)),
                ('start_order', otree.db.models.PositiveIntegerField(null=True, db_index=True)),
                ('label', otree.db.models.CharField(max_length=50, null=True)),
                ('_index_in_subsessions', otree.db.models.PositiveIntegerField(null=True, default=0)),
                ('_index_in_pages', otree.db.models.PositiveIntegerField(null=True, db_index=True, default=0)),
                ('id_in_session', otree.db.models.PositiveIntegerField(null=True)),
                ('_waiting_for_ids', otree.db.models.CharField(max_length=300, null=True)),
                ('code', otree.db.models.CharField(max_length=16, unique=True, null=True, default=otree.common_internal.random_chars_8)),
                ('last_request_succeeded', otree.db.models.BooleanField(verbose_name='Health of last server request', choices=[(True, 'Yes'), (False, 'No')])),
                ('visited', otree.db.models.BooleanField(db_index=True, default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('ip_address', otree.db.models.GenericIPAddressField(null=True)),
                ('_last_page_timestamp', otree.db.models.PositiveIntegerField(null=True)),
                ('_last_request_timestamp', otree.db.models.PositiveIntegerField(null=True)),
                ('is_on_wait_page', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('_current_page_name', otree.db.models.CharField(verbose_name='page', max_length=200, null=True)),
                ('_current_app_name', otree.db.models.CharField(verbose_name='app', max_length=200, null=True)),
                ('_round_number', otree.db.models.PositiveIntegerField(null=True)),
                ('_current_form_page_url', otree.db.models.URLField(null=True)),
                ('_max_page_index', otree.db.models.PositiveIntegerField(null=True)),
                ('_is_auto_playing', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
            ],
            options={
                'ordering': ['pk'],
            },
            bases=(otree.models.varsmixin._SaveTheChangeWithCustomFieldSupport, models.Model),
        ),
        migrations.CreateModel(
            name='ParticipantLockModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('participant_code', models.CharField(max_length=16, unique=True)),
                ('locked', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='ParticipantRoomVisit',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('room_name', models.CharField(max_length=50)),
                ('participant_label', models.CharField(max_length=200)),
                ('tab_unique_id', models.CharField(max_length=20, unique=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='ParticipantToPlayerLookup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('page_index', models.PositiveIntegerField()),
                ('app_name', models.CharField(max_length=300)),
                ('player_pk', models.PositiveIntegerField()),
                ('url', models.CharField(max_length=300)),
                ('participant', models.ForeignKey(to='otree.Participant')),
            ],
        ),
        migrations.CreateModel(
            name='RoomToSession',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('room_name', models.CharField(max_length=255, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('vars', otree.db.models._JSONField(null=True, default=dict)),
                ('config', otree.db.models._JSONField(null=True, default=dict)),
                ('label', otree.db.models.CharField(max_length=300, blank=True, null=True, help_text='For internal record-keeping')),
                ('experimenter_name', otree.db.models.CharField(max_length=300, blank=True, null=True, help_text='For internal record-keeping')),
                ('code', otree.db.models.CharField(max_length=16, unique=True, null=True, default=otree.common_internal.random_chars_8)),
                ('time_scheduled', otree.db.models.DateTimeField(blank=True, null=True, help_text='For internal record-keeping')),
                ('time_started', otree.db.models.DateTimeField(null=True)),
                ('mturk_HITId', otree.db.models.CharField(max_length=300, blank=True, null=True, help_text='Hit id for this session on MTurk')),
                ('mturk_HITGroupId', otree.db.models.CharField(max_length=300, blank=True, null=True, help_text='Hit id for this session on MTurk')),
                ('mturk_qualification_type_id', otree.db.models.CharField(max_length=300, blank=True, null=True, help_text='Qualification type that is assigned to each worker taking hit')),
                ('mturk_num_participants', otree.db.models.IntegerField(null=True, default=-1, help_text='Number of participants on MTurk')),
                ('mturk_sandbox', otree.db.models.BooleanField(default=True, choices=[(True, 'Yes'), (False, 'No')], help_text='Should this session be created in mturk sandbox?')),
                ('archived', otree.db.models.BooleanField(db_index=True, default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('comment', otree.db.models.TextField(blank=True, null=True)),
                ('_anonymous_code', otree.db.models.CharField(max_length=10, null=True, db_index=True, default=otree.common_internal.random_chars_10)),
                ('_pre_create_id', otree.db.models.CharField(max_length=300, null=True, db_index=True)),
                ('_use_browser_bots', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('_bots_errored', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('_bots_finished', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('_cannot_restart_bots', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('has_bots', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('is_demo', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
                ('ready', otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')])),
            ],
            options={
                'ordering': ['pk'],
            },
            bases=(otree.models.varsmixin._SaveTheChangeWithCustomFieldSupport, models.Model),
        ),
        migrations.CreateModel(
            name='UndefinedFormModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
            ],
        ),
        migrations.AddField(
            model_name='roomtosession',
            name='session',
            field=models.ForeignKey(to='otree.Session'),
        ),
        migrations.AddField(
            model_name='participant',
            name='session',
            field=models.ForeignKey(to='otree.Session'),
        ),
        migrations.AddField(
            model_name='pagetimeout',
            name='participant',
            field=models.ForeignKey(to='otree.Participant'),
        ),
        migrations.AddField(
            model_name='pagecompletion',
            name='participant',
            field=models.ForeignKey(to='otree.Participant'),
        ),
        migrations.AddField(
            model_name='pagecompletion',
            name='session',
            field=models.ForeignKey(to='otree.Session'),
        ),
        migrations.AlterUniqueTogether(
            name='expectedroomparticipant',
            unique_together=set([('room_name', 'participant_label')]),
        ),
        migrations.AddField(
            model_name='completedsubsessionwaitpage',
            name='session',
            field=models.ForeignKey(to='otree.Session'),
        ),
        migrations.AddField(
            model_name='completedgroupwaitpage',
            name='session',
            field=models.ForeignKey(to='otree.Session'),
        ),
        migrations.AlterUniqueTogether(
            name='participanttoplayerlookup',
            unique_together=set([('participant', 'page_index')]),
        ),
        migrations.AlterIndexTogether(
            name='participanttoplayerlookup',
            index_together=set([('participant', 'page_index')]),
        ),
        migrations.AlterIndexTogether(
            name='participant',
            index_together=set([('session', 'mturk_worker_id', 'mturk_assignment_id')]),
        ),
        migrations.AlterIndexTogether(
            name='pagetimeout',
            index_together=set([('participant', 'page_index')]),
        ),
        migrations.AlterUniqueTogether(
            name='completedsubsessionwaitpage',
            unique_together=set([('page_index', 'session')]),
        ),
        migrations.AlterIndexTogether(
            name='completedsubsessionwaitpage',
            index_together=set([('page_index', 'session')]),
        ),
        migrations.AlterUniqueTogether(
            name='completedgroupwaitpage',
            unique_together=set([('page_index', 'session', 'group_pk')]),
        ),
        migrations.AlterIndexTogether(
            name='completedgroupwaitpage',
            index_together=set([('page_index', 'session', 'group_pk')]),
        ),
        migrations.AddField(
            model_name='participant',
            name='_browser_bot_finished',
            field=otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')]),
        ),
        migrations.CreateModel(
            name='GlobalLockModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('locked', models.BooleanField(default=False)),
            ],
        ),
        migrations.RemoveField(
            model_name='participant',
            name='_is_auto_playing',
        ),
        migrations.AddField(
            model_name='participant',
            name='_is_bot',
            field=otree.db.models.BooleanField(default=False, choices=[(True, 'Yes'), (False, 'No')]),
        ),
        migrations.AlterModelOptions(
            name='participant',
            options={},
        ),
        migrations.AlterModelOptions(
            name='session',
            options={},
        ),
        migrations.AlterIndexTogether(
            name='participant',
            index_together=set([]),
        ),
        migrations.CreateModel(
            name='BrowserBotsLauncherSessionCode',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, serialize=False, auto_created=True)),
                ('code', models.CharField(max_length=10)),
                ('is_only_record', models.BooleanField(unique=True, default=True)),
            ],
        ),
        migrations.AddField(
            model_name='session',
            name='_bot_case_number',
            field=otree.db.models.PositiveIntegerField(null=True),
        ),
        migrations.RenameField(
            model_name='session',
            old_name='_use_browser_bots',
            new_name='use_browser_bots',
        ),
        migrations.AlterModelOptions(
            name='participant',
            options={'ordering': ['pk']},
        ),
        migrations.AlterModelOptions(
            name='session',
            options={'ordering': ['pk']},
        ),
        migrations.AlterIndexTogether(
            name='participant',
            index_together=set([('session', 'mturk_worker_id', 'mturk_assignment_id')]),
        ),
        migrations.RenameField(
            model_name='completedgroupwaitpage',
            old_name='after_all_players_arrive_run',
            new_name='fully_completed',
        ),
        migrations.RenameField(
            model_name='completedsubsessionwaitpage',
            old_name='after_all_players_arrive_run',
            new_name='fully_completed',
        ),
        migrations.AlterField(
            model_name='session',
            name='_pre_create_id',
            field=otree.db.models.CharField(max_length=255, null=True, db_index=True),
        ),
        migrations.AddField(
            model_name='completedgroupwaitpage',
            name='id_in_subsession',
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name='session',
            name='_admin_report_app_names',
            field=otree.db.models.TextField(null=True, default=''),
        ),
        migrations.AddField(
            model_name='session',
            name='_admin_report_num_rounds',
            field=otree.db.models.CharField(max_length=255, null=True, default=''),
        ),
        migrations.AlterUniqueTogether(
            name='completedgroupwaitpage',
            unique_together=set([('page_index', 'session', 'id_in_subsession')]),
        ),
        migrations.AlterIndexTogether(
            name='completedgroupwaitpage',
            index_together=set([('page_index', 'session', 'id_in_subsession')]),
        ),
        migrations.RemoveField(
            model_name='completedgroupwaitpage',
            name='group_pk',
        ),
        migrations.RemoveField(
            model_name='session',
            name='time_scheduled',
        ),
        migrations.RemoveField(
            model_name='session',
            name='time_started',
        ),
    ]



# -*- coding: utf-8 -*-
# <standard imports>
from __future__ import division
from otree.api import (
    BaseSubsession, BaseGroup, BasePlayer, BaseConstants, models
)
# </standard imports>

doc = "foo"

class Constants(BaseConstants):
    name_in_url = 'advance_slowest_2'
    players_per_group = 2
    num_rounds = 1


class Subsession(BaseSubsession):
    pass


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import otree.db.models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0006_auto_20160708_0657'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='browserbotsubmit',
            name='participant',
        ),
        migrations.RemoveField(
            model_name='browserbotsubmit',
            name='session',
        ),
        migrations.RemoveField(
            model_name='participant',
            name='_is_auto_playing',
        ),
        migrations.RemoveField(
            model_name='session',
            name='special_category',
        ),
        migrations.AddField(
            model_name='participant',
            name='_is_bot',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='_bots_errored',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='_bots_finished',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='_cannot_restart_bots',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='has_bots',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='is_demo',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.AddField(
            model_name='session',
            name='ready',
            field=otree.db.models.BooleanField(choices=[(True, 'Yes'), (False, 'No')], default=False),
        ),
        migrations.DeleteModel(
            name='BrowserBotSubmit',
        ),
    ]



#!/usr/bin/env python
import os
import sys


if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

    from otree.management.cli import execute_from_command_line
    execute_from_command_line(sys.argv, script_file=__file__)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import otree.db.models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0009_browserbotslaunchersessioncode'),
    ]

    operations = [
        migrations.AddField(
            model_name='session',
            name='_bot_case_number',
            field=otree.db.models.PositiveIntegerField(null=True),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from django.conf import settings
from django.test.utils import override_settings

import otree
from otree import apps

from .base import TestCase


class TestApps(TestCase):

    @override_settings(RAVEN_CONFIG={})
    def test_patch_raven_config(self):
        apps.patch_raven_config()
        expected = {
            'release': '{}{}'.format(
                otree.get_version(), ',dbg' if settings.DEBUG else '')}
        self.assertEquals(settings.RAVEN_CONFIG, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import division

import six
from django.db.models import Prefetch
from otree_save_the_change.mixins import SaveTheChange
from otree.db import models
from otree.common_internal import get_models_module
from otree import matching
import copy

ATTRIBUTE_ERROR_MESSAGE = '''
Subsession object has no attribute '{}'. If it is a model field or method,
it must be declared on the Subsession class in models.py.
'''.replace('\n', '')


class BaseSubsession(SaveTheChange, models.Model):
    """Base class for all Subsessions.
    """

    class Meta:
        abstract = True
        ordering = ['pk']
        index_together = ['session', 'round_number']

    session = models.ForeignKey(
        'otree.Session', related_name='%(app_label)s_%(class)s', null=True)

    round_number = models.PositiveIntegerField(
        db_index=True,
        doc='''If this subsession is repeated (i.e. has multiple rounds), this
        field stores the position (index) of this subsession, among subsessions
        in the same app.

        For example, if a session consists of the subsessions:

            [app1, app2, app1, app1, app3]

        Then the round numbers of these subsessions would be:

            [1, 1, 2, 3, 1]

        '''
    )

    def __getattribute__(self, name):
        try:
            return super(BaseSubsession, self).__getattribute__(name)
        except AttributeError:
            raise AttributeError(ATTRIBUTE_ERROR_MESSAGE.format(name)) from None

    def in_rounds(self, first, last):
        qs = type(self).objects.filter(
            session=self.session,
            round_number__range=(first, last),
        ).order_by('round_number')

        return list(qs)

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number-1)

    def in_all_rounds(self):
        return self.in_previous_rounds() + [self]

    def name(self):
        return str(self.pk)

    def __unicode__(self):
        return self.name()

    @property
    def app_name(self):
        return self._meta.app_config.name

    def in_round(self, round_number):
        return type(self).objects.get(
            session=self.session,
            round_number=round_number
        )

    def _get_players_per_group_list(self):
        """get a list whose elements are the number of players in each group

        Example: a group of 30 players

        # everyone is in the same group
        [30]

        # 5 groups of 6
        [6, 6, 6, 6, 6,]

        # 2 groups of 5 players, 2 groups of 10 players
        [5, 10, 5, 10] # (you can do this with players_per_group = [5, 10]

        """

        ppg = self._Constants.players_per_group
        subsession_size = self.player_set.count()
        if ppg is None:
            return [subsession_size]

        # if groups have variable sizes, you can put it in a list
        if isinstance(ppg, (list, tuple)):
            assert all(n > 1 for n in ppg)
            group_cycle = ppg
        else:
            assert isinstance(ppg, six.integer_types) and ppg > 1
            group_cycle = [ppg]

        num_group_cycles = subsession_size // sum(group_cycle)
        return group_cycle * num_group_cycles

    def get_groups(self):
        return list(self.group_set.order_by('id_in_subsession'))

    def get_players(self):
        return list(self.player_set.order_by('pk'))

    def get_group_matrix(self):
        players_prefetch = Prefetch(
            'player_set',
            queryset=self._PlayerClass().objects.order_by('id_in_group'),
            to_attr='_ordered_players')
        return [group._ordered_players
                for group in self.group_set.order_by('id_in_subsession')
                                 .prefetch_related(players_prefetch)]

    def set_group_matrix(self, matrix):
        """
        warning: this deletes the groups and any data stored on them
        TODO: we should indicate this in docs
        """

        try:
            players_flat = [p for g in matrix for p in g]
        except TypeError:
            raise TypeError(
                'Group matrix must be a list of lists.'
            ) from None
        try:
            matrix_pks = sorted(p.pk for p in players_flat)
        except AttributeError:
            # if integers, it's OK
            if isinstance(players_flat[0], int):
                # deep copy so that we don't modify the input arg
                matrix = copy.deepcopy(matrix)
                players_flat = sorted(players_flat)
                players_from_db = self.get_players()
                if players_flat == list(range(1, len(players_from_db) + 1)):
                    for i, row in enumerate(matrix):
                        for j, val in enumerate(row):
                            matrix[i][j] = players_from_db[val - 1]
                else:
                    raise ValueError(
                        'If you pass a matrix of integers to this function, '
                        'It must contain all integers from 1 to '
                        'the number of players in the subsession.'
                    ) from None
            else:
                raise TypeError(
                    'The elements of the group matrix '
                    'must either be Player objects, or integers.'
                ) from None

        else:
            existing_pks = list(
                self.player_set.values_list(
                    'pk', flat=True
                ).order_by('pk'))
            if matrix_pks != existing_pks:
                raise ValueError(
                    'The group matrix must contain each player '
                    'in the subsession exactly once.'
                )

        # Before deleting groups, Need to set the foreignkeys to None
        # 2016-11-8: does this need to be in a transaction?
        # because what if a player refreshes their page while this is going
        # on?
        self.player_set.update(group=None)
        self.group_set.all().delete()

        GroupClass = self._GroupClass()
        for i, row in enumerate(matrix, start=1):
            group = GroupClass.objects.create(
                subsession=self, id_in_subsession=i,
                session=self.session, round_number=self.round_number)

            group.set_players(row)

    def set_groups(self, matrix):
        '''renamed this to set_group_matrix, but keeping in for compat'''
        return self.set_group_matrix(matrix)

    def check_group_integrity(self):
        ''' should be moved from here to a test case'''
        players = self.player_set.values_list('pk', flat=True)
        players_from_groups = self.player_set.filter(
            group__subsession=self).values_list('pk', flat=True)
        if not set(players) == set(players_from_groups):
            raise Exception('Group integrity check failed')

    @property
    def _Constants(self):
        return get_models_module(self._meta.app_config.name).Constants

    def _GroupClass(self):
        return models.get_model(self._meta.app_config.label, 'Group')

    def _PlayerClass(self):
        return models.get_model(self._meta.app_config.label, 'Player')

    def _has_group_by_arrival_time(self):
        from otree.common_internal import has_group_by_arrival_time
        return has_group_by_arrival_time(self._meta.app_config.name)

    def _first_round_group_matrix(self):
        players = list(self.get_players())

        ppg = self._Constants.players_per_group
        if ppg is None or self._has_group_by_arrival_time():
            ppg = len(players)

        group_matrix = []
        for i in range(0, len(players), ppg):
            group_matrix.append(players[i:i + ppg])
        return group_matrix

    def _create_groups(self):
        if self.round_number == 1:
            group_matrix = self._first_round_group_matrix()
            self.set_group_matrix(group_matrix)
        else:
            self.group_like_round(1)

    def group_like_round(self, round_number):
        previous_round = self.in_round(round_number)
        group_matrix = [
            group._ordered_players
            for group in previous_round.group_set.order_by(
                'id_in_subsession'
            ).prefetch_related(
                Prefetch(
                    'player_set',
                    queryset=self._PlayerClass().objects.order_by(
                        'id_in_group'),
                    to_attr='_ordered_players'
                )
            )
        ]
        for i, group_list in enumerate(group_matrix):
            for j, player in enumerate(group_list):
                # for every entry (i,j) in the matrix, follow the pointer
                # to the same person in the next round
                group_matrix[i][j] = player.in_round(self.round_number)

        self.set_group_matrix(group_matrix)

    def group_randomly(self, *, fixed_id_in_group=False):
        group_matrix = self.get_group_matrix()
        group_matrix = matching.randomly(
            group_matrix,
            fixed_id_in_group)
        self.set_group_matrix(group_matrix)

    def group_by_rank(self, ranked_list):
        group_matrix = matching.by_rank(
            ranked_list,
            self._Constants.players_per_group
        )
        self.set_group_matrix(group_matrix)

    def before_session_starts(self):
        '''Deprecated and renamed to creating_session'''
        pass

    def creating_session(self):
        pass

    def vars_for_admin_report(self):
        return {}


# -*- coding: utf-8 -*-
# <standard imports>
from __future__ import division
from otree.api import (
    BaseSubsession, BaseGroup, BasePlayer, BaseConstants, models
)
# </standard imports>

doc = "foo"

class Constants(BaseConstants):
    name_in_url = 'admin_report'
    players_per_group = None
    num_rounds = 2


class Subsession(BaseSubsession):
    foo = models.PositiveIntegerField(initial=42)

    def vars_for_admin_report(self):
        return {'custom_var': 43}


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    pass



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants

class MyPage(Page):

    form_fields = ['my_field']
    form_model = models.Player

class ResultsWaitPage(WaitPage):
    pass

class Results(Page):
    pass

page_sequence = [
        MyPage,
        ResultsWaitPage,
        Results
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six


class BaseConstantsMeta(type):

    def __setattr__(cls, attr, value):
        raise AttributeError("Constants are read-only.")


class BaseConstants(six.with_metaclass(BaseConstantsMeta, object)):

    pass



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import views
from .models import Constants
from otree.api import Bot


class PlayerBot(Bot):

    def play_round(self):
        # add these in later
        assert self.group.set_in_wait_page == True, self.group.id
        assert self.player.set_in_wait_page == True

        yield views.Page1

        # Race condition: this might be evaluated before all players finish
        # currently it works, but if i change the order that bots run,
        # it might not work in the future.
        # for now, this is the simplest way to get it working with browser bots
        # because there is no validate_play method. (usually its not needed)
        num_players = self.subsession.player_set.count()
        if self.player.id_in_subsession == num_players and self.subsession.round_number == Constants.num_rounds:
            for subsession in self.subsession.in_all_rounds():
                groups = subsession.get_groups()
                players = subsession.get_players()
                for group in groups:
                    assert len(group.get_players()) == Constants.players_per_group, self.subsession.round_number
                assert len(groups) == len(players) / Constants.players_per_group




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import sys
from functools import reduce
from collections import OrderedDict
from decimal import Decimal
import warnings

import six
from six.moves import range
from six.moves import zip

from django.conf import settings
from django.db import transaction
from django.db.utils import OperationalError

import schema

import otree.db.idmap
from otree.models import Participant, Session
from otree.deprecate import OtreeDeprecationWarning
from otree.common_internal import (
    get_models_module, get_app_constants, validate_alphanumeric,
    min_players_multiple, get_bots_module)
import otree.common_internal
from otree.common import RealWorldCurrency
from otree.models_concrete import ParticipantLockModel
import otree.bots.browser


def gcd(a, b):
    """Return greatest common divisor using Euclid's Algorithm."""
    while b:
        a, b = b, a % b
    return a


def lcm(a, b):
    """Return lowest common multiple."""
    return a * b // gcd(a, b)


def lcmm(*args):
    """Return lcm of args."""
    return reduce(lcm, args)


class SessionConfig(dict):

    def get_lcm(self):
        min_multiple_list = []
        for app_name in self['app_sequence']:
            app_constants = get_app_constants(app_name)
            # if players_per_group is None, 0, etc.
            min_multiple = min_players_multiple(
                app_constants.players_per_group)
            min_multiple_list.append(min_multiple)
        return lcmm(*min_multiple_list)

    def get_num_bot_cases(self):
        num_cases = 1
        for app_name in self['app_sequence']:
            bots_module = get_bots_module(app_name)
            cases = bots_module.PlayerBot.cases
            num_cases = max(num_cases, len(cases))
        return num_cases

    def clean(self):

        config_schema = schema.Schema({
            'name': str,
            'app_sequence': list,
            'participation_fee': object,
            'real_world_currency_per_point': object,
            'num_demo_participants': int,
            'doc': str,
            str: object,
        })

        try:
            config_schema.validate(self)
        except schema.SchemaError as e:
            raise ValueError('settings.SESSION_CONFIGS: {}'.format(e)) from None

        # Allow non-ASCII chars in session config keys, because they are
        # configurable in the admin, so need to be readable by non-English
        # speakers. However, don't allow punctuation, spaces, etc.
        # They make it harder to reason about and could cause problems
        # later on. also could uglify the user's code.

        INVALID_IDENTIFIER_MSG = (
            'Key "{}" in settings.SESSION_CONFIGS '
            'must not contain spaces, punctuation, '
            'or other special characters. '
            'It can contain non-English characters, '
            'but it must be a valid Python variable name '
            'according to string.isidentifier().'
        )

        for key in self:
            if not key.isidentifier():
                raise ValueError(INVALID_IDENTIFIER_MSG.format(key))

        validate_alphanumeric(
            self['name'],
            identifier_description='settings.SESSION_CONFIGS name'
        )

        app_sequence = self['app_sequence']
        if len(app_sequence) != len(set(app_sequence)):
            msg = (
                'settings.SESSION_CONFIGS: '
                'app_sequence of "{}" '
                'must not contain duplicate elements. '
                'If you want multiple rounds, '
                'you should set Constants.num_rounds.')
            raise ValueError(msg.format(self['name']))

        if len(app_sequence) == 0:
            raise ValueError(
                'settings.SESSION_CONFIGS: Need at least one subsession.')

        self.setdefault('display_name', self['name'])
        self.setdefault('doc', '')

        # TODO: fixed_pay is deprecated as of 2015-05-07,
        # in favor of participation_fee. make this required at some point.
        if (('participation_fee' not in self) and
                ('fixed_pay' in self)):
            warn_msg = (
                "'fixed_pay' is deprecated; "
                "you should rename it to 'participation_fee'.")
            warnings.warn(warn_msg, OtreeDeprecationWarning)

            self['participation_fee'] = self['fixed_pay']

        self['participation_fee'] = RealWorldCurrency(
            self['participation_fee'])

    def app_sequence_display(self):
        app_sequence = []
        for app_name in self['app_sequence']:
            models_module = get_models_module(app_name)
            num_rounds = models_module.Constants.num_rounds
            formatted_app_name = otree.common_internal.app_name_format(
                app_name)
            if num_rounds > 1:
                formatted_app_name = '{} ({} rounds)'.format(
                    formatted_app_name, num_rounds)
            subsssn = {
                'doc': getattr(models_module, 'doc', ''),
                'name': formatted_app_name}
            app_sequence.append(subsssn)
        return app_sequence

    non_editable_fields = {
        'app_sequence',
        'name',
        'display_name',
        'app_sequence',
        'num_demo_participants',
        'doc',
        'num_bots',
    }

    def builtin_editable_fields(self):
        fields = ['participation_fee']
        if settings.USE_POINTS:
            fields.append('real_world_currency_per_point')
        return fields

    def custom_editable_fields(self):
        # should there also be some restriction on
        # what chars are allowed? because maybe not all chars work
        # in an HTML form field (e.g. periods, quotes, etc)
        # so far, it seems any char works OK, even without escaping
        # before making an HTML attribute. even '>漢 ."&'
        # so i'll just put a general recommendation in the docs

        fields = [
            k for k, v in self.items()
            if k not in self.non_editable_fields
            and k not in self.builtin_editable_fields()
            and type(v) in [bool, int, float, str]]

        # they're in a dict so we can't preserve the original ordering
        # this is the best we can do
        return sorted(fields)

    def editable_fields(self):
        return self.builtin_editable_fields() + self.custom_editable_fields()

    def html_field_name(self, field_name):
        return '{}.{}'.format(self['name'], field_name)

    def editable_field_html(self, field_name):
        existing_value = self[field_name]
        html_field_name = self.html_field_name(field_name)
        base_attrs = ["name='{}'".format(html_field_name)]

        if isinstance(existing_value, bool):
            attrs = [
                "type='checkbox'",
                'checked' if existing_value else '',
                # don't use class=form-control because it looks too big,
                # like it's intended for mobile devices
            ]
        elif isinstance(existing_value, int):
            attrs = [
                "type='number'",
                "required",
                "step='1'",
                "value='{}'".format(existing_value),
                "class='form-control'",
            ]
        elif isinstance(existing_value, (float, Decimal)):
            # convert to float, e.g. participation_fee
            attrs = [
                "class='form-control'",
                "type='number'",
                "step='any'",
                "required",
                "value='{}'".format(float(existing_value)),
            ]
        elif isinstance(existing_value, str):
            attrs = [
                "type='text'",
                "value='{}'".format(existing_value),
                "class='form-control'"
            ]
        html = '''
        <tr><td><b>{}</b><td><input {}></td>
        '''.format(field_name, ' '.join(base_attrs + attrs))
        return html

    def builtin_editable_fields_html(self):
        return [self.editable_field_html(k)
                for k in self.builtin_editable_fields()]

    def custom_editable_fields_html(self):
        return [self.editable_field_html(k)
                for k in self.custom_editable_fields()]


def get_session_configs_dict():
    SESSION_CONFIGS_DICT = OrderedDict()
    for config_dict in settings.SESSION_CONFIGS:
        config_obj = SessionConfig(settings.SESSION_CONFIG_DEFAULTS)
        config_obj.update(config_dict)
        config_obj.clean()
        SESSION_CONFIGS_DICT[config_dict['name']] = config_obj
    return SESSION_CONFIGS_DICT

SESSION_CONFIGS_DICT = get_session_configs_dict()


def app_labels_from_sessions(config_names):
    apps = set()
    for config_name in config_names:
        config = SESSION_CONFIGS_DICT[config_name]
        apps.update(config["app_sequence"])
    return apps


def create_session(
        session_config_name, label='', num_participants=None,
        _pre_create_id=None,
        room_name=None, for_mturk=False, use_cli_bots=False,
        is_demo=False, force_browser_bots=False,
        honor_browser_bots_config=False, bot_case_number=None,
        edited_session_config_fields=None):

    session = None
    use_browser_bots = False
    participants = []
    edited_session_config_fields = edited_session_config_fields or {}

    with transaction.atomic():
        # 2014-5-2: i could implement this by overriding the __init__ on the
        # Session model, but I don't really know how that works, and it seems
        # to be a bit discouraged: http://goo.gl/dEXZpv
        # 2014-9-22: preassign to groups for demo mode.

        otree.db.idmap.activate_cache()

        try:
            session_config = SESSION_CONFIGS_DICT[session_config_name]
        except KeyError:
            msg = 'Session config "{}" not found in settings.SESSION_CONFIGS.'
            raise KeyError(msg.format(session_config_name)) from None
        else:
            # copy so that we don't mutate the original
            # .copy() returns a dict, so need to convert back to SessionConfig
            session_config = SessionConfig(session_config.copy())
            session_config.update(edited_session_config_fields)

            # check validity and converts serialized decimal & currency values
            # back to their original data type (because they were serialized
            # when passed through channels
            session_config.clean()

        if force_browser_bots:
            use_browser_bots = True
        elif (session_config.get('use_browser_bots') and
              honor_browser_bots_config):
            use_browser_bots = True
        else:
            use_browser_bots = False
        if use_browser_bots and bot_case_number is None:
            # choose one randomly
            num_bot_cases = session_config.get_num_bot_cases()
            # choose bot case number randomly...maybe reconsider this?
            # we can only run one.
            bot_case_number = random.choice(range(num_bot_cases))

        session = Session.objects.create(
            config=session_config,
            label=label,
            _pre_create_id=_pre_create_id,
            use_browser_bots=use_browser_bots,
            is_demo=is_demo,
            _bot_case_number=bot_case_number) # type: Session

        def bulk_create(model, descriptions):
            model.objects.bulk_create([
                model(session=session, **description)
                for description in descriptions])
            return model.objects.filter(session=session).order_by('pk')

        # check that it divides evenly
        session_lcm = session_config.get_lcm()
        if num_participants % session_lcm:
            msg = (
                'Session Config {}: Number of participants ({}) does not '
                'divide evenly into group size ({})'
            ).format(session_config['name'], num_participants, session_lcm)
            raise ValueError(msg)

        if for_mturk:
            session.mturk_num_participants = (
                    num_participants /
                    settings.MTURK_NUM_PARTICIPANTS_MULTIPLE)

        start_order = list(range(num_participants))
        if session_config.get('random_start_order'):
            random.shuffle(start_order)

        participants = bulk_create(
            Participant,
            [{
                'id_in_session': id_in_session,
                'start_order': j,
                # check if id_in_session is in the bots ID list
                '_is_bot': use_cli_bots or use_browser_bots,
             }
             for id_in_session, j in enumerate(start_order, start=1)])

        ParticipantLockModel.objects.bulk_create([
            ParticipantLockModel(participant_code=participant.code)
            for participant in participants])

        for app_name in session_config['app_sequence']:

            models_module = get_models_module(app_name)
            app_constants = get_app_constants(app_name)

            round_numbers = list(range(1, app_constants.num_rounds + 1))

            subs = bulk_create(
                models_module.Subsession,
                [{'round_number': round_number}
                 for round_number in round_numbers])

            # Create players
            models_module.Player.objects.bulk_create([
                models_module.Player(
                    session=session,
                    subsession=subsession,
                    round_number=round_number,
                    participant=participant)
                for round_number, subsession in zip(round_numbers, subs)
                for participant in participants])

        session._create_groups_and_initialize()

        session.build_participant_to_player_lookups()
        # automatically save all objects since the cache was activated:
        # Player, Group, Subsession, Participant, Session
        otree.db.idmap.save_objects()
        otree.db.idmap.deactivate_cache()

    if use_browser_bots:
        # what about clear_browser_bots? if session is created through
        # UI, when do we run that? it should be run when the session
        # is deleted
        try:
            for participant in participants:
                otree.bots.browser.initialize_bot(
                    participant.code)
        except:
            session.delete()
            raise

    session._set_admin_report_app_names()
    session.ready = True
    session.save()

    # this should happen after session.ready = True
    if room_name is not None:
        from otree.room import ROOM_DICT
        room = ROOM_DICT[room_name]
        room.session = session

    return session


default_app_config = 'otree.session.apps.OtreeSessionConfig'



from __future__ import division

from otree.api import Bot, SubmissionMustFail

from . import views
from .models import Constants
from django.conf import settings
from collections import namedtuple

LocStrings = namedtuple(
    'LocStrings',
    ['form_has_errors', 'next', 'yes', 'time_left', 'user_defined_string'])


STRINGS = {
    'de': LocStrings(
        form_has_errors='Bitte korrigieren',
        next='Weiter',
        yes='Ja',
        time_left='Verbleibende Zeit',
        user_defined_string='Fußball'
    ),
    'zh-hans': LocStrings(
        form_has_errors='请修复',
        next='下一页',
        yes='是',
        time_left='时间',
        user_defined_string='足球'
    ),
}

class PlayerBot(Bot):

    def play_round(self):
        strings = STRINGS[settings.LANGUAGE_CODE]
        assert strings.next in self.html
        assert strings.yes in self.html
        assert strings.time_left in self.html
        assert strings.user_defined_string in self.html
        yield SubmissionMustFail(views.Page1)
        assert strings.form_has_errors in self.html



# -*- coding: utf-8 -*-
from __future__ import division

from otree.api import Bot

from . import views
from .models import Constants


class PlayerBot(Bot):

    def play_round(self):
        if self.subsession.round_number == Constants.num_rounds:
            yield (views.MyPage)
            yield (views.Results)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# deprecation shim
import warnings

from otree.deprecate import OtreeDeprecationWarning

from django.core.management.base import BaseCommand


DEPRECATION_STRING = ('celery command no longer exists in oTree 0.5+. '
                      'You should update your Procfile.')


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        warnings.warn(DEPRECATION_STRING, OtreeDeprecationWarning)



import floppyforms

import otree.forms
import otree.widgets
from otree.forms import ModelForm

from .base import TestCase
from .models import FormFieldModel


class TestModelForm(ModelForm):

    class Meta:
        model = FormFieldModel
        exclude = ()


class UseFloppyformWidgetsTests(TestCase):

    def test_overriden_django_fields(self):
        self.assertIsInstance(
            TestModelForm.base_fields['char'], floppyforms.CharField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['boolean'],
            floppyforms.TypedChoiceField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['big_integer'], floppyforms.IntegerField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['char'], floppyforms.CharField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['comma_separated_integer'],
            floppyforms.CharField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['date'], floppyforms.DateField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['date_time'], floppyforms.DateTimeField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['decimal'], floppyforms.DecimalField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['email'], floppyforms.EmailField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['file'], floppyforms.FileField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['file_path'],
            floppyforms.FilePathField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['float'], floppyforms.FloatField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['integer'], floppyforms.IntegerField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['generic_ip_address'],
            floppyforms.GenericIPAddressField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['positive_integer'],
            floppyforms.IntegerField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['positive_small_integer'],
            floppyforms.IntegerField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['slug'], floppyforms.SlugField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['small_integer'],
            floppyforms.IntegerField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['text'], floppyforms.CharField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['time'], floppyforms.TimeField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['url'], floppyforms.URLField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['one_to_one'],
            floppyforms.ModelChoiceField
        )

    def test_custom_fields(self):
        self.assertIsInstance(
            TestModelForm.base_fields['sent_amount'],
            otree.forms.CurrencyChoiceField
        )

    def test_currency_field(self):
        self.assertIsInstance(
            TestModelForm.base_fields['currency'],
            otree.forms.CurrencyField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['currency'].widget,
            otree.widgets._CurrencyInput
        )
        self.assertIsInstance(
            TestModelForm.base_fields['currency_choice'],
            otree.forms.CurrencyChoiceField
        )
        self.assertIsInstance(
            TestModelForm.base_fields['currency_choice'].widget,
            otree.widgets.Select
        )



#!/usr/bin/env python
# encoding: utf-8

import warnings
import datetime
from collections import defaultdict
import sys
import logging

from django.conf import settings
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, HttpResponseServerError
from django.shortcuts import get_object_or_404

from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse

import vanilla

import boto.mturk.connection
from boto.mturk.connection import MTurkRequestError
from boto.mturk.qualification import Qualifications

import IPy

import otree
from otree import forms
from otree.deprecate import OtreeDeprecationWarning
from otree.views.abstract import AdminSessionPageMixin
from otree.checks.mturk import validate_session_for_mturk
from otree.forms import widgets
from otree.common import RealWorldCurrency
from otree.models import Session


logger = logging.getLogger('otree')


class MTurkError(Exception):

    def __init__(self, request, message):
        self.message = message
        messages.error(request, message, extra_tags='safe')

    def __str__(self):
        return self.message


class MTurkConnection(boto.mturk.connection.MTurkConnection):

    def __init__(self, request, in_sandbox=True):
        if in_sandbox:
            self.mturk_host = settings.MTURK_SANDBOX_HOST
        else:
            self.mturk_host = settings.MTURK_HOST
        self.request = request

    def __enter__(self):
        self = boto.mturk.connection.MTurkConnection(
            aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
            host=self.mturk_host,
        )
        return self

    def __exit__(self, exc_type, value, traceback):
        # TODO: need to take care of possible errors (login,
        # "service not approved")
        if exc_type is MTurkRequestError:
            MTurkError(self.request, value.message)
        return False


def get_workers_by_status(conn, hit_id):
    all_assignments = get_all_assignments(conn, hit_id)
    workers_by_status = defaultdict(list)
    for assignment in all_assignments:
        workers_by_status[
            assignment.AssignmentStatus
        ].append(assignment.WorkerId)
    return workers_by_status


def get_all_assignments(conn, hit_id, status=None):
    # Accumulate all relevant assignments, one page of results at
    # a time.
    assignments = []
    page = 1
    while True:
        rs = conn.get_assignments(
            hit_id=hit_id,
            page_size=100,
            page_number=page,
            status=status)
        assignments.extend(rs)
        if len(assignments) >= int(rs.TotalNumResults):
            break
        page += 1
    return assignments


class MTurkCreateHITForm(forms.Form):

    in_sandbox = forms.BooleanField(
        required=False,
        help_text="Do you want HIT published on MTurk sandbox?")
    title = forms.CharField()
    description = forms.CharField()
    keywords = forms.CharField()
    money_reward = forms.RealWorldCurrencyField(
        # it seems that if this is omitted, the step defaults to an integer,
        # meaninng fractional inputs are not accepted
        widget=widgets._RealWorldCurrencyInput(attrs={'step': 0.01})
    )
    assignments = forms.IntegerField(
        label="Number of assignments",
        help_text="How many unique Workers do you want to work on the HIT?")
    minutes_allotted_per_assignment = forms.IntegerField(
        label="Minutes allotted per assignment",
        help_text=(
            "Number of minutes, that a Worker has to "
            "complete the HIT after accepting it."
        ))
    expiration_hours = forms.FloatField(
        label="Hours until HIT expiration",
        help_text=(
            "Number of hours after which the HIT "
            "is no longer available for users to accept. "
        ))

    def __init__(self, *args, **kwargs):
        super(MTurkCreateHITForm, self).__init__(*args, **kwargs)
        self.fields['assignments'].widget.attrs['readonly'] = True


class MTurkCreateHIT(AdminSessionPageMixin, vanilla.FormView):
    '''This view creates mturk HIT for session provided in request
    AWS externalQuestion API is used to generate HIT.

    '''
    form_class = MTurkCreateHITForm

    def in_public_domain(self, request, *args, **kwargs):
        """This method validates if oTree are published on a public domain
        because mturk need it

        """
        host = request.get_host().lower()
        if ":" in host:
            host = host.split(":", 1)[0]
        if host == "localhost":
            return False
        try:
            ip = IPy.IP(host)
            return ip.iptype() == "PUBLIC"
        except ValueError:
            # probably is a public domain
            return True

    def get(self, request, *args, **kwargs):
        validate_session_for_mturk(request, self.session)
        mturk_settings = self.session.config['mturk_hit_settings']
        initial = {
            'title': mturk_settings['title'],
            'description': mturk_settings['description'],
            'keywords': ', '.join(mturk_settings['keywords']),
            'money_reward': self.session.config['participation_fee'],
            'in_sandbox': settings.DEBUG,
            'minutes_allotted_per_assignment': (
                mturk_settings['minutes_allotted_per_assignment']
            ),
            'expiration_hours': mturk_settings['expiration_hours'],
            'assignments': self.session.mturk_num_participants,
        }
        form = self.get_form(initial=initial)
        context = self.get_context_data(form=form)
        context['mturk_enabled'] = (
            bool(settings.AWS_ACCESS_KEY_ID) and
            bool(settings.AWS_SECRET_ACCESS_KEY)
        )
        context['runserver'] = 'runserver' in sys.argv
        url = self.request.build_absolute_uri(
            reverse('MTurkCreateHIT', args=(self.session.code,))
        )
        secured_url = urlunparse(urlparse(url)._replace(scheme='https'))
        context['secured_url'] = secured_url

        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        form = self.get_form(
            data=request.POST,
            files=request.FILES
        )
        if not form.is_valid():
            return self.form_invalid(form)
        session = self.session
        in_sandbox = 'in_sandbox' in form.data
        # session can't be created
        if (not self.in_public_domain(request, *args, **kwargs) and
           not in_sandbox):
                msg = (
                    '<h1>Error: '
                    'oTree must run on a public domain for Mechanical Turk'
                    '</h1>')
                return HttpResponseServerError(msg)
        with MTurkConnection(self.request, in_sandbox) as mturk_connection:
            mturk_settings = session.config['mturk_hit_settings']
            qualification_id = mturk_settings.get(
                'grant_qualification_id', None)
            # verify that specified qualification type
            # for preventing retakes exists on mturk server
            if qualification_id:
                try:
                    mturk_connection.get_qualification_type(qualification_id)
                except MTurkRequestError as e:
                    code = 'AWS.MechanicalTurk.QualificationTypeDoesNotExist'
                    if e.error_code == code:
                        if in_sandbox:
                            sandbox_note = (
                                'You are currently using the sandbox, so you '
                                'can only grant qualifications that were '
                                'also created in the sandbox.')
                        else:
                            sandbox_note = (
                                'You are using the MTurk live site, so you '
                                'can only grant qualifications that were '
                                'also created on the live site, and not the '
                                'MTurk sandbox.')
                        msg = (
                            "In settings.py you specified qualification ID "
                            " '{}' which doesn't exist in your MTurk account. "
                            "Please check its validity. Note: {}".format(
                                qualification_id, sandbox_note))
                        if in_sandbox:
                            msg += ' You are currently'
                        messages.error(request, msg)
                        return HttpResponseRedirect(
                            reverse(
                                'MTurkCreateHIT', args=(session.code,)))
                else:
                    session.mturk_qualification_type_id = qualification_id

            url_landing_page = self.request.build_absolute_uri(
                reverse('MTurkLandingPage', args=(session.code,)))

            # updating schema from http to https
            # this is compulsory for MTurk exteranlQuestion
            # TODO: validate, that the server support https
            #       (heroku does support by default)
            secured_url_landing_page = urlunparse(
                urlparse(url_landing_page)._replace(scheme='https'))

            # TODO: validate that there is enought money for the hit
            money_reward = form.data['money_reward']
            reward = boto.mturk.price.Price(
                amount=float(money_reward))

            # assign back to participation_fee, in case it was changed
            # in the form
            # TODO: why do I have to explicitly
            # convert back to RealWorldCurrency?
            # shouldn't it already be that?
            session.config['participation_fee'] = RealWorldCurrency(
                money_reward)

            # creating external questions, that would be passed to the hit
            external_question = boto.mturk.question.ExternalQuestion(
                secured_url_landing_page, mturk_settings['frame_height'])

            qualifications = mturk_settings.get('qualification_requirements')

            # deprecated summer 2015: remove this
            if not qualifications and \
               hasattr(settings, 'MTURK_WORKER_REQUIREMENTS'):
                    warn_msg = (
                        'The MTURK_WORKER_REQUIREMENTS setting has been '
                        'deprecated. You should instead use '
                        '"qualification_requirements" as shown here: '
                        'https://github.com/oTree-org/oTree/blob/master/'
                        'settings.py')
                    warnings.warn(warn_msg, OtreeDeprecationWarning)
                    qualifications = settings.MTURK_WORKER_REQUIREMENTS

            mturk_hit_parameters = {
                'title': form.cleaned_data['title'],
                'description': form.cleaned_data['description'],
                'keywords': [
                    k.strip() for k in
                    form.cleaned_data['keywords'].split(',')
                ],
                'question': external_question,
                'max_assignments': form.cleaned_data['assignments'],
                'reward': reward,
                'response_groups': ('Minimal', 'HITDetail'),
                'qualifications': Qualifications(qualifications),
                'duration': datetime.timedelta(
                    minutes=(
                        form.cleaned_data['minutes_allotted_per_assignment']
                    )),
                'lifetime': datetime.timedelta(
                    minutes=int(60*form.cleaned_data['expiration_hours']))
            }

            hit = mturk_connection.create_hit(**mturk_hit_parameters)
            session.mturk_HITId = hit[0].HITId
            session.mturk_HITGroupId = hit[0].HITGroupId
            session.mturk_sandbox = in_sandbox
            session.save()

        return HttpResponseRedirect(
            reverse('MTurkCreateHIT', args=(session.code,)))


class PayMTurk(vanilla.View):

    url_pattern = r'^PayMTurk/(?P<session_code>[a-z0-9]+)/$'

    def post(self, request, *args, **kwargs):
        session = get_object_or_404(otree.models.Session,
                                    code=kwargs['session_code'])
        successful_payments = 0
        failed_payments = 0
        with MTurkConnection(self.request,
                             session.mturk_sandbox) as mturk_connection:
            for p in session.participant_set.filter(
                mturk_assignment_id__in=request.POST.getlist('payment')
            ):
                try:
                    # approve assignment
                    mturk_connection.approve_assignment(p.mturk_assignment_id)
                except boto.mturk.connection.MTurkRequestError as e:
                    msg = (
                        'Could not pay {} because of an error communicating '
                        'with MTurk: {}'.format(p._id_in_session(), str(e)))
                    messages.error(request, msg)
                    logger.error(msg)
                    failed_payments += 1
                else:
                    successful_payments += 1
                    # grant bonus
                    # TODO: check if bonus was already paid before,
                    # perhaps through
                    # mturk requester webinterface
                    payoff = p.payoff_in_real_world_currency().to_number()
                    if payoff > 0:
                        bonus = boto.mturk.price.Price(amount=payoff)
                        mturk_connection.grant_bonus(
                            p.mturk_worker_id,
                            p.mturk_assignment_id,
                            bonus,
                            reason="Thank you for "
                            "participating.")
        msg = 'Successfully made {} payments.'.format(successful_payments)
        if failed_payments > 0:
            msg += ' {} payments failed.'.format(failed_payments)
            messages.warning(request, msg)
        else:
            messages.success(request, msg)
        return HttpResponseRedirect(
            reverse('MTurkSessionPayments', args=(session.code,)))


class RejectMTurk(vanilla.View):

    url_pattern = r'^RejectMTurk/(?P<session_code>[a-z0-9]+)/$'

    def post(self, request, *args, **kwargs):
        session = get_object_or_404(Session,
                                    code=kwargs['session_code'])
        with MTurkConnection(self.request,
                             session.mturk_sandbox) as mturk_connection:
            for p in session.participant_set.filter(
                mturk_assignment_id__in=request.POST.getlist('payment')
            ):
                mturk_connection.reject_assignment(p.mturk_assignment_id)

        messages.success(request, "You successfully rejected "
                                  "selected assignments")
        return HttpResponseRedirect(
            reverse('MTurkSessionPayments', args=(session.code,)))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0019_delete_expectedroomparticipant'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pagetimeout',
            name='expiration_time',
            field=models.FloatField(),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0013_failedsessioncreation_traceback'),
    ]

    operations = [
        migrations.RenameField(
            model_name='completedgroupwaitpage',
            old_name='after_all_players_arrive_run',
            new_name='fully_completed',
        ),
        migrations.RenameField(
            model_name='completedsubsessionwaitpage',
            old_name='after_all_players_arrive_run',
            new_name='fully_completed',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import models
from otree.api import WaitPage
from tests.utils import BlankTemplatePage as Page
from .models import Constants


class MyWait(WaitPage):
    group_by_arrival_time = True

    def get_players_for_group(self, waiting_players):
        a_players = [p for p in waiting_players if p.type == 'A']
        b_players = [p for p in waiting_players if p.type == 'B']

        new_group = a_players[:1] + b_players[:1]
        if len(new_group) == 2:
            return new_group

    def is_displayed(self):
        return self.round_number == 1

class Page1(Page):
    pass


page_sequence = [
    MyWait,
    Page1,
]



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import
from . import views
from otree.api import Bot


class PlayerBot(Bot):

    def play_round(self):
        yield views.Page1
        yield views.Page2



from typing import Any, List
from otree.common import Currency, RealWorldCurrency

## This code is duplicated in several places
# bots/__init__.pyi, views/__init__.pyi, models/__init__.pyi
# importing doesn't seem to work with PyCharm autocomplete

class Session:

    config = None  # type: dict
    vars = None  # type: dict
    def get_participants(self) -> List['Participant']: pass
    def get_subsessions(self) -> List['Subsession']: pass

class Participant:

    session = None # type: Session
    vars = None  # type: dict
    label = None  # type: str
    id_in_session = None  # type: int
    payoff = None  # type: Currency

    def get_players(self) -> List['Player']: pass
    def payoff_plus_participation_fee(self) -> RealWorldCurrency: pass


class Bot:
    html = '' # type: str
    case = None # type: Any
    cases = [] # type: List
    participant = None  # type: Participant
    session = None # type: Participant

def Submission(PageClass, post_data: dict={}, *, check_html=True, timeout_happened=False): pass
def SubmissionMustFail(PageClass, post_data: dict={}, *, check_html=True): pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings

from django.conf import settings
import django.test

# =============================================================================
# HELPER
# =============================================================================


# 2016-06-16: is this still needed? TODO
class OTreeTestClient(django.test.client.Client):

    def login(self):
        return super(OTreeTestClient, self).login(
            username=settings.ADMIN_USERNAME, password=settings.ADMIN_PASSWORD)


class TestCase(django.test.TestCase):

    client_class = OTreeTestClient

    def assertWarns(self, warning, callable, *args, **kwds):
        with warnings.catch_warnings(record=True) as warning_list:
            warnings.simplefilter('always')

            callable(*args, **kwds)

        condition = any(item.category == warning for item in warning_list)
        msg = "'{}' not warned".format(str(warning))
        self.assertTrue(condition, msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from django.core.urlresolvers import reverse
import django.test.client
import unittest
from otree.session import create_session
from .utils import get_path


def has_header(response, header_name):
    return header_name in response


def is_wait_page(response):
    return has_header(response, 'oTree-Wait-Page')


def get_html(response):
    return response.content.decode('utf-8')


def participant_initialized(response):
    redirect_chain = [ele[0] for ele in response.redirect_chain]
    return any('InitializeParticipant' in url for url in redirect_chain)


class RoomTestCase(unittest.TestCase):
    '''Subclassed from unittest.TestCase,
    because we don't want to flush the participant labels from the DB
    after each test, because it circumvents room._participant_labels_loaded.
    '''

    def setUp(self):
        self.browser = django.test.client.Client()

    def get(self, url):
        resp = self.browser.get(url, follow=True)
        self.assertEqual(resp.status_code, 200)
        self.path = get_path(resp, url)
        return resp

    def post(self, url, data=None):
        data = data or {}
        resp = self.browser.post(url, data, follow=True)
        self.assertEqual(resp.status_code, 200)
        self.path = get_path(resp, url)
        return resp


class TestRoomWithoutSession(RoomTestCase):

    def setUp(self):
        self.browser = django.test.client.Client()

    def test_open_admin_links(self):
        urls = [
            reverse('Rooms'),
            reverse('RoomWithoutSession', args=['default']),
            reverse('RoomWithoutSession', args=['anon']),
        ]

        for url in urls:
            self.get(url)

    def test_open_participant_links(self):

        room_name = 'default'
        room_with_labels = reverse('AssignVisitorToRoom', args=[room_name])

        resp = self.get(room_with_labels)
        self.assertIn('Please enter your participant label', get_html(resp))

        resp = self.get(
            room_with_labels + '?participant_label=JohnSmith',
        )

        self.assertIn('room/{}'.format(room_name), self.path)

        resp = self.browser.get(
            room_with_labels + '?participant_label=NotInLabelsFile',
        )
        self.assertEqual(resp.status_code, 404)

        anon_base_url = reverse('AssignVisitorToRoom', args=['anon'])

        resp = self.get(anon_base_url)
        self.assertIn('Waiting for your session to begin', get_html(resp))

        resp = self.get(anon_base_url + '?participant_label=JohnSmith')
        self.assertIn('Waiting for your session to begin', get_html(resp))

    def test_ping(self):
        pass


class TestRoomWithSession(RoomTestCase):
    def setUp(self):
        self.browser = django.test.client.Client()
        self.default_session = create_session(
            'simple',
            # make it 6 so that we can test if the participant is reassigned
            # if they open their start link again (1/6 chance)
            num_participants=6,
            room_name='default',
        )

        self.anon_session = create_session(
            'simple',
            num_participants=2,
            room_name='anon',
        )

    def test_open_admin_links(self):
        self.get(reverse('RoomWithoutSession', args=['default']))
        self.assertTrue('room_with_session' in self.path)

    def test_open_participant_links(self):
        room_with_labels = reverse('AssignVisitorToRoom', args=['default'])

        resp = self.get(room_with_labels)
        self.assertIn('Please enter your participant label', get_html(resp))

        resp = self.get(
            room_with_labels + '?participant_label=JohnSmith',
        )
        self.assertTrue(participant_initialized(resp))

        resp = self.browser.get(
            room_with_labels + '?participant_label=NotInLabelsFile',
        )
        self.assertEqual(resp.status_code, 404)

        anon_base_url = reverse('AssignVisitorToRoom', args=['anon'])

        resp = self.get(anon_base_url)
        self.assertTrue(participant_initialized(resp))

        resp = self.get(anon_base_url + '?participant_label=JohnSmith')
        self.assertTrue(participant_initialized(resp))

    def test_participant_label(self):
        room_with_labels = reverse('AssignVisitorToRoom', args=['default'])

        participant_label = 'JohnSmith'
        session_participant_codes = [
            p.code for p in self.default_session.get_participants()]
        assigned_participants = []

        # make sure reopening the same link assigns you to same participant
        for i in range(2):
            resp = self.get(
                room_with_labels + '?participant_label={}'.format(
                    participant_label))

            # last URL in redirect chain, then [0] (i forget why)
            page_url = resp.redirect_chain[-1][0]

            for code in session_participant_codes:
                if code in page_url:
                    assigned_participants.append(code)
                    break
        self.assertEqual(assigned_participants[0], assigned_participants[1])
        participant = self.default_session.participant_set.get(
            code=assigned_participants[0])

        # make sure participant_label is recorded
        self.assertEqual(participant.label, participant_label)


    def test_close_room(self):

        url = reverse('CloseRoom', args=['default'])
        self.get(url)

        self.get(reverse('RoomWithoutSession', args=['default']))
        self.assertTrue('room_without_session' in self.path)

    def test_delete_session_in_room(self):
        pass

    def test_session_start_links_room(self):
        pass

    def tearDown(self):
        url = reverse('CloseRoom', args=['default'])
        self.get(url)

        url = reverse('CloseRoom', args=['anon'])
        self.get(url)



# -*- coding: utf-8 -*-
from __future__ import division

from otree.api import Bot

from . import views
from .models import Constants


class PlayerBot(Bot):

    def play_round(self):
        yield (views.Page1, {'my_field': 3})



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0010_session__bot_case_number'),
    ]

    operations = [
        migrations.RenameField(
            model_name='session',
            old_name='_use_browser_bots',
            new_name='use_browser_bots',
        ),
    ]



# -*- coding: utf-8 -*-
# <standard imports>
from __future__ import division
from otree.db import models
import otree.models
# </standard imports>


doc = """
Test misc functionality of a 3-player game
"""


class Constants:
    name_in_url = 'misc_3p'
    players_per_group = 3
    num_rounds = 20


class Subsession(otree.models.BaseSubsession):

    def creating_session(self):
        if self.round_number == 2:
            for group in self.get_groups():
                players = group.get_players()
                pks = [p.pk for p in players]
                for i, p in enumerate(players):
                    assert p.id_in_group == i + 1
                players.reverse()
                group.set_players(players)
                pks_reversed = list(reversed(pks))
                assert [p.pk for p in group.get_players()] == pks_reversed

    def shuffle_p1(self):
        matrix = [g.get_players() for g in self.get_groups()]
        p1_list = [g_list[0] for g_list in matrix]
        p1_in_first_group = p1_list.pop(0)
        p1_list.append(p1_in_first_group)
        for i, p_list in enumerate(matrix):
            matrix[i][0] = p1_list[i]
        self.set_group_matrix(matrix)


class Group(otree.models.BaseGroup):

    def set_payoffs(self):
        for p in self.get_players():
            p.payoff = 50

    in_all_groups_wait_page = models.FloatField(initial=0)


class Player(otree.models.BasePlayer):

    def other_player(self):
        """Returns other player in group. Only valid for 2-player groups."""
        return self.get_others_in_group()[0]

    from_other_player = models.PositiveIntegerField()

    is_winner = models.BooleanField(initial=False)
    in_all_groups_wait_page = models.FloatField(initial=0)

    group_id_before_p1_switch = models.PositiveIntegerField()

    def role(self):
        # you can make this depend of self.id_in_group
        return ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from typing import List
from otree.common import RealWorldCurrency

## This code is duplicated in several places
# bots/__init__.pyi, views/__init__.pyi, models/__init__.pyi
# importing doesn't seem to work with PyCharm autocomplete

class Session:

    config = None  # type: dict
    vars = None  # type: dict
    def get_participants(self) -> List['Participant']: pass
    def get_subsessions(self) -> List['Subsession']: pass

class Participant:

    session = None # type: Session
    vars = None  # type: dict
    label = None  # type: str
    id_in_session = None  # type: int
    payoff = None  # type: Currency

    def get_players(self) -> List['Player']: pass
    def payoff_plus_participation_fee(self) -> RealWorldCurrency: pass


class WaitPage:
    wait_for_all_groups = False
    group_by_arrival_time = False
    title_text = None
    body_text = None
    template_name = None
    round_number = None  # type: int
    participant = None  # type: Participant
    session = None  # type: Session

    def is_displayed(self) -> bool: pass
    def after_all_players_arrive(self): pass
    def get_players_for_group(self, waiting_players): pass


class Page:
    round_number = None  # type: int
    template_name = None # type: str
    timeout_seconds = None # type: int
    timeout_submission = None # type: dict
    timeout_happened = None # type: bool
    timer_text = None # type: str
    participant = None  # type: Participant
    session = None  # type: Session
    form_model = None #
    form_fields = None  # type: List[str]

    def get_form_fields(self) -> List['str']: pass
    def vars_for_template(self) -> dict: pass
    def before_next_page(self): pass
    def is_displayed(self) -> bool: pass
    def error_message(self, values): pass
    def get_timeout_seconds(self): pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import otree.db.models


class Migration(migrations.Migration):

    dependencies = [
        ('otree', '0015_auto_20161111_1055'),
    ]

    operations = [
        migrations.AddField(
            model_name='completedgroupwaitpage',
            name='id_in_subsession',
            field=models.PositiveIntegerField(default=0),
        ),
        migrations.AddField(
            model_name='session',
            name='_admin_report_app_names',
            field=otree.db.models.TextField(default='', null=True),
        ),
        migrations.AddField(
            model_name='session',
            name='_admin_report_num_rounds',
            field=otree.db.models.CharField(default='', max_length=255, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='completedgroupwaitpage',
            unique_together=set([('page_index', 'session', 'id_in_subsession')]),
        ),
        migrations.AlterIndexTogether(
            name='completedgroupwaitpage',
            index_together=set([('page_index', 'session', 'id_in_subsession')]),
        ),
        migrations.RemoveField(
            model_name='completedgroupwaitpage',
            name='group_pk',
        ),
    ]



from otree.session import create_session
from otree.models.participant import Participant
from otree.bots.bot import ParticipantBot
from .base import TestCase
from tests.timeout_submission.models import Player, Constants
from tests.timeout_submission import views
import django.test
from otree.api import Submission, Currency, Page
import itertools
import time
import otree.timeout.tasks #.submit_expired_url.schedule(
from unittest.mock import patch, MagicMock

test_client = django.test.Client()

# tuple instead of dict so we don't mutate it by mistake
default_submission = (
    ('f_bool', True),
    ('f_char', 'hello'),
    ('f_currency', Currency(2)),
    ('f_float', 0.1),
    ('f_posint', 3),
)


class PageWithTimeout(Page):

    a = 0

    def get_timeout_seconds(self):
        PageWithTimeout.a += 1
        return PageWithTimeout.a


class PageWithNoTimeout(Page):
    pass


class TestTimeout(TestCase):

    def get_page(self, PageClass, participant):
        page = PageClass()
        page.set_attributes(participant)
        page.request = MagicMock()
        page.request.path = 'foo'
        return page

    def setUp(self):
        session = create_session(
            session_config_name='timeout_submission',
            num_participants=1,
            use_cli_bots=True,
        )
        self.participant = session.get_participants()[0]

        # simulate opening the start link in the most minimal way
        # a more heavyweight approach would be to create the ParticipantBot
        # and open the start URL
        self.participant._index_in_pages = 1
        self.participant.save()

    def test_page_refresh(self):
        '''
        Test that when you refresh a page, oTree still remembers the original
        timeout, rather creating a new one.
        '''

        page = self.get_page(PageWithTimeout, self.participant)
        expiration_time_1 = time.time() + page.remaining_timeout_seconds()

        page = self.get_page(PageWithTimeout, self.participant)
        expiration_time_2 = time.time() + page.remaining_timeout_seconds()

        # if you call it a second time, it should be the same
        self.assertEqual(expiration_time_1, expiration_time_2)

    def test_timeout_scheduling(self):
        '''Loading a page with timeout should schedule a page submission'''

        with patch.object(otree.timeout.tasks.submit_expired_url, 'schedule') as schedule_method:
            page = self.get_page(PageWithNoTimeout, self.participant)
            page.remaining_timeout_seconds()

            self.assertFalse(schedule_method.called)

            page = self.get_page(PageWithTimeout, self.participant)
            page.remaining_timeout_seconds()

            self.assertTrue(schedule_method.called)

        # calling remaining_timeout_seconds() twice in the same request should
        # not schedule a second time. Testing this because the template calls
        # {{ view.remaining_timeout_seconds }} several times
        with patch.object(otree.timeout.tasks.submit_expired_url, 'schedule') as schedule_method:
            page.remaining_timeout_seconds()
            self.assertFalse(schedule_method.called)


class TestTimeoutSubmission(TestCase):

    def setUp(self):
        create_session(
            session_config_name='timeout_submission',
            num_participants=1,
            use_cli_bots=True,
        )

    def submit_form(self, values, timeout_happened):
        participant = Participant.objects.get()
        bot = ParticipantBot(participant, load_player_bots=False)
        bot.open_start_url()
        bot.submit(Submission(views.Page1, values, timeout_happened=timeout_happened))

    def timeout_submit_form(self, values):
        self.submit_form(values, timeout_happened=True)

    def assert_player_fields(self, values):
        player = Player.objects.get()
        for field_name, value in values.items():
            self.assertEqual(getattr(player, field_name), value, msg=field_name)

    def test_no_timeout(self):
        '''baseline test'''
        values = dict(default_submission)
        self.submit_form(values, timeout_happened=False)

        player = Player.objects.get()
        self.assertEqual(player.timeout_happened, False)

    def test_timeout_on_regular_page(self):
        participant = Participant.objects.get()
        bot = ParticipantBot(participant, load_player_bots=False)
        bot.open_start_url()
        bot.submit(Submission(views.Page1, {}, timeout_happened=True))
        # it should be OK to use timeout_happened=True, even if the page
        # has no timeout_seconds, because you can be simulating "advance slowest"
        bot.submit(Submission(views.PageWithoutTimeout, {}, timeout_happened=True))

    def test_valid(self):
        '''valid form'''
        values = dict(default_submission)

        self.timeout_submit_form(values)
        self.assert_player_fields(dict(default_submission))

        # test that timeout_happened was set
        player = Player.objects.get()
        self.assertEqual(player.timeout_happened, True)

    def test_invalid_fields(self):
        '''invalid individual fields but passes error_message'''
        values = dict(default_submission)
        values.pop('f_float')
        values['f_posint'] = -1

        self.timeout_submit_form(values)

        values = dict(default_submission)
        values['f_float'] = 0
        values['f_posint'] = 0

        self.assert_player_fields(values)

    def test_valid_but_invalid_error_message(self):
        '''valid individual fields but passes error_message'''
        values = dict(default_submission)
        values['f_char'] = Constants.invalid_f_char

        self.timeout_submit_form(values)

        auto_submit_defaults = {
            'f_bool': False,
            'f_char': '',
            'f_currency': Currency(0),
            'f_float': 0,
            'f_posint': 0,
        }

        self.assert_player_fields(auto_submit_defaults)

    def test_invalid_and_invalid_error_message(self):
        '''invalid individual fields and invalid error_message'''
        values = dict(default_submission)
        values['f_char'] = Constants.invalid_f_char
        values['f_posint'] = -1

        self.timeout_submit_form(values)

        auto_submit_defaults = {
            'f_bool': False,
            'f_char': '',
            'f_currency': Currency(0),
            'f_float': 0,
            'f_posint': 0,
        }

        self.assert_player_fields(auto_submit_defaults)
